import dataclasses
import logging
from typing import ClassVar

import numpy as np
import torch

from .annrescaler import AnnRescalerDet
from .. import headmeta
from ..visualizer import CifDet as CifDetVisualizer
from ..utils import create_sink, mask_valid_area

LOG = logging.getLogger(__name__)


@dataclasses.dataclass
class CifDet:
    meta: headmeta.CifDet
    rescaler: AnnRescalerDet = None
    v_threshold: int = 0
    bmin: float = 1.0  #: in pixels
    visualizer: CifDetVisualizer = None

    side_length: ClassVar[int] = 5
    padding: ClassVar[int] = 10

    def __call__(self, image, anns, meta):
        return CifDetGenerator(self)(image, anns, meta)


class CifDetGenerator():
    def __init__(self, config: CifDet):
        self.config = config

        self.rescaler = config.rescaler or AnnRescalerDet(
            config.meta.stride, len(config.meta.categories))
        self.visualizer = config.visualizer or CifDetVisualizer(config.meta)

        self.intensities = None
        self.fields_reg = None
        self.fields_wh = None
        self.fields_reg_bmin = None
        self.fields_wh_bmin = None
        self.fields_reg_l = None

        self.sink = create_sink(config.side_length)
        self.s_offset = (config.side_length - 1.0) / 2.0

    def __call__(self, image, anns, meta):
        width_height_original = image.shape[2:0:-1]

        detections = self.rescaler.detections(anns)
        bg_mask = self.rescaler.bg_mask(anns, width_height_original,
                                        crowd_margin=(self.config.side_length - 1) / 2)
        valid_area = self.rescaler.valid_area(meta)
        LOG.debug('valid area: %s, pif side length = %d', valid_area, self.config.side_length)

        n_fields = len(self.config.meta.categories)
        self.init_fields(n_fields, bg_mask)
        self.fill(detections)
        fields = self.fields(valid_area)

        self.visualizer.processed_image(image)
        self.visualizer.targets(fields, annotation_dicts=anns)

        return fields

    def init_fields(self, n_fields, bg_mask):
        field_w = bg_mask.shape[-1] + 2 * self.config.padding
        field_h = bg_mask.shape[-2] + 2 * self.config.padding
        self.intensities = np.zeros((n_fields, field_h, field_w), dtype=np.float32)
        self.fields_reg = np.full((n_fields, 2, field_h, field_w), np.nan, dtype=np.float32)
        self.fields_wh = np.full((n_fields, 2, field_h, field_w), np.nan, dtype=np.float32)
        self.fields_reg_bmin = np.full((n_fields, field_h, field_w), np.nan, dtype=np.float32)
        self.fields_wh_bmin = np.full((n_fields, field_h, field_w), np.nan, dtype=np.float32)
        self.fields_reg_l = np.full((n_fields, field_h, field_w), np.inf, dtype=np.float32)

        # bg_mask
        p = self.config.padding
        self.fields_reg_l[:, p:-p, p:-p][bg_mask == 0] = 1.0
        self.intensities[:, p:-p, p:-p][bg_mask == 0] = np.nan

    def fill(self, detections):
        for category_id, bbox in detections:
            xy = bbox[:2] + 0.5 * bbox[2:]
            wh = bbox[2:]
            self.fill_detection(category_id - 1, xy, wh)

    def fill_detection(self, f, xy, wh):
        ij = np.round(xy - self.s_offset).astype(np.int) + self.config.padding
        minx, miny = int(ij[0]), int(ij[1])
        maxx, maxy = minx + self.config.side_length, miny + self.config.side_length
        if minx < 0 or maxx > self.intensities.shape[2] or \
           miny < 0 or maxy > self.intensities.shape[1]:
            return

        offset = xy - (ij + self.s_offset - self.config.padding)
        offset = offset.reshape(2, 1, 1)

        # mask
        sink_reg = self.sink + offset
        sink_l = np.linalg.norm(sink_reg, axis=0)
        mask = sink_l < self.fields_reg_l[f, miny:maxy, minx:maxx]
        core_radius = (self.config.side_length - 1) / 2.0
        mask_fringe = np.logical_and(
            sink_l > core_radius,
            sink_l < self.fields_reg_l[f, miny:maxy, minx:maxx],
        )
        self.fields_reg_l[f, miny:maxy, minx:maxx][mask] = sink_l[mask]

        # update intensity
        self.intensities[f, miny:maxy, minx:maxx][mask] = 1.0
        self.intensities[f, miny:maxy, minx:maxx][mask_fringe] = np.nan

        # update regression
        self.fields_reg[f, :, miny:maxy, minx:maxx][:, mask] = sink_reg[:, mask]

        # update wh
        assert wh[0] > 0.0
        assert wh[1] > 0.0
        self.fields_wh[f, :, miny:maxy, minx:maxx][:, mask] = np.expand_dims(wh, 1)

        # update bmin
        half_scale = 0.5 * min(wh[0], wh[1])
        bmin = max(0.1 * half_scale, self.config.bmin / self.config.meta.stride)
        self.fields_reg_bmin[f, miny:maxy, minx:maxx][mask] = bmin
        self.fields_wh_bmin[f, miny:maxy, minx:maxx][mask] = bmin

    def fields(self, valid_area):
        p = self.config.padding
        intensities = self.intensities[:, p:-p, p:-p]
        fields_reg = self.fields_reg[:, :, p:-p, p:-p]
        fields_wh = self.fields_wh[:, :, p:-p, p:-p]
        fields_reg_bmin = self.fields_reg_bmin[:, p:-p, p:-p]
        fields_wh_bmin = self.fields_wh_bmin[:, p:-p, p:-p]

        mask_valid_area(intensities, valid_area)
        mask_valid_area(fields_reg[:, 0], valid_area, fill_value=np.nan)
        mask_valid_area(fields_reg[:, 1], valid_area, fill_value=np.nan)
        mask_valid_area(fields_wh[:, 0], valid_area, fill_value=np.nan)
        mask_valid_area(fields_wh[:, 1], valid_area, fill_value=np.nan)
        mask_valid_area(fields_reg_bmin, valid_area, fill_value=np.nan)
        mask_valid_area(fields_wh_bmin, valid_area, fill_value=np.nan)

        return torch.from_numpy(np.concatenate([
            np.expand_dims(intensities, 1),
            fields_reg,
            fields_wh,
            np.expand_dims(fields_reg_bmin, 1),
            np.expand_dims(fields_wh_bmin, 1),
        ], axis=1))



import io
import logging

import numpy as np
import PIL
import torch

from .preprocess import Preprocess

try:
    import scipy
except ImportError:
    scipy = None

LOG = logging.getLogger(__name__)


class ImageTransform(Preprocess):
    """Transform image without modifying annotations or meta."""
    def __init__(self, image_transform):
        self.image_transform = image_transform

    def __call__(self, image, anns, meta):
        image = self.image_transform(image)
        return image, anns, meta


class JpegCompression(Preprocess):
    """Add jpeg compression."""
    def __init__(self, quality=50):
        self.quality = quality

    def __call__(self, image, anns, meta):
        f = io.BytesIO()
        image.save(f, 'jpeg', quality=self.quality)
        return PIL.Image.open(f), anns, meta


class Blur(Preprocess):
    """Blur image."""
    def __init__(self, max_sigma=5.0):
        self.max_sigma = max_sigma

    def __call__(self, image, anns, meta):
        im_np = np.asarray(image)
        sigma = self.max_sigma * float(torch.rand(1).item())
        im_np = scipy.ndimage.filters.gaussian_filter(im_np, sigma=(sigma, sigma, 0))
        return PIL.Image.fromarray(im_np), anns, meta


class HorizontalBlur(Preprocess):
    def __init__(self, sigma=5.0):
        self.sigma = sigma

    def __call__(self, image, anns, meta):
        im_np = np.asarray(image)
        sigma = self.sigma * (0.8 + 0.4 * float(torch.rand(1).item()))
        LOG.debug('horizontal blur with %f', sigma)
        im_np = scipy.ndimage.filters.gaussian_filter1d(im_np, sigma=sigma, axis=1)
        return PIL.Image.fromarray(im_np), anns, meta



import copy
import logging
import random

import PIL

from ..preprocess import Preprocess

LOG = logging.getLogger(__name__)


class BlankPast(Preprocess):
    def __call__(self, images, all_anns, metas):
        all_anns = copy.deepcopy(all_anns)
        metas = copy.deepcopy(metas)

        for i, _ in enumerate(images[1:], start=1):
            images[i] = PIL.Image.new('RGB', (320, 240), (127, 127, 127))

        for i, _ in enumerate(all_anns[1:], start=1):
            all_anns[i] = []

        for meta in metas[1:]:
            meta['image'] = {'frame_id': -1, 'file_name': 'blank'}
            assert 'annotations' not in meta

        return images, all_anns, metas


class PreviousPast(Preprocess):
    def __init__(self):
        self.previous_image = PIL.Image.new('RGB', (320, 240), (127, 127, 127))
        self.previous_meta = {'frame_id': -1, 'file_name': 'blank'}
        self.previous_annotations = []

    def __call__(self, images, all_anns, metas):
        all_anns = copy.deepcopy(all_anns)
        metas = copy.deepcopy(metas)

        LOG.debug('replacing %s with %s', metas[1], self.previous_meta)

        for i, _ in enumerate(images[1:], start=1):
            images[i] = self.previous_image

        for i, _ in enumerate(all_anns[1:], start=1):
            all_anns[i] = []  # TODO assumes previous image has nothing to do with current

        for meta in metas[1:]:
            meta['image'] = self.previous_meta  # why image?
            assert 'annotations' not in meta  # why would there be anns in meta?

        self.previous_image = images[0]
        self.previous_annotations = all_anns[0]
        self.previous_meta = metas[0]
        return images, all_anns, metas


class RandomizeOneFrame(Preprocess):
    def __init__(self):
        self.previous_image = None
        self.previous_meta = None
        self.previous_annotations = []

    def __call__(self, images, all_anns, metas):
        all_anns = copy.deepcopy(all_anns)
        metas = copy.deepcopy(metas)

        replace_index = random.randrange(0, len(metas))

        if self.previous_image is not None:
            # image
            images[replace_index] = self.previous_image

            # annotations
            all_anns[replace_index] = self.previous_annotations
            if self.previous_meta.get('annotation_file', 0) \
               == metas[replace_index].get('annotation_file', 1):
                pass
            else:
                for ann in all_anns[replace_index]:
                    ann['track_id'] += 10000

            # meta
            metas[replace_index] = self.previous_meta

        not_replaced_index = 0 if replace_index != 0 else 1
        self.previous_image = copy.deepcopy(images[not_replaced_index])
        self.previous_annotations = copy.deepcopy(all_anns[not_replaced_index])
        self.previous_meta = copy.deepcopy(metas[not_replaced_index])
        return images, all_anns, metas



import openpifpaf

from .cocokpst import CocoKpSt
from .posetrack2018 import Posetrack2018
from .posetrack2017 import Posetrack2017


def register():
    openpifpaf.CHECKPOINT_URLS['tshufflenetv2k16'] = openpifpaf.PRETRAINED_UNAVAILABLE
    openpifpaf.CHECKPOINT_URLS['tshufflenetv2k30'] = (
        'http://github.com/openpifpaf/torchhub/releases/download/v0.12.10/'
        'tshufflenetv2k30-210628-075118-posetrack2018-cocokpst-slurm668247-o25-3d734bb8.pkl')

    openpifpaf.DATAMODULES['posetrack2018'] = Posetrack2018
    openpifpaf.DATAMODULES['posetrack2017'] = Posetrack2017
    openpifpaf.DATAMODULES['cocokpst'] = CocoKpSt



import openpifpaf

from . import datamodule
from . import basenet


def register():
    openpifpaf.DATAMODULES['cifar10'] = datamodule.Cifar10

    openpifpaf.BASE_TYPES.add(basenet.Cifar10Net)
    openpifpaf.BASE_FACTORIES['cifar10net'] = lambda: basenet.Cifar10Net()  # pylint: disable=unnecessary-lambda



import copy
import logging

from .base import Base
from ..annotation import AnnotationCrowd, AnnotationDet
from .. import headmeta, show

try:
    import matplotlib.cm
    CMAP_GREENS_NAN = copy.copy(matplotlib.cm.get_cmap('Greens'))
    CMAP_GREENS_NAN.set_bad('white', alpha=0.5)
except ImportError:
    CMAP_GREENS_NAN = None

LOG = logging.getLogger(__name__)


class CifDet(Base):
    """Visualize a CifDet field."""

    def __init__(self, meta: headmeta.CifDet):
        super().__init__(meta.name)
        self.meta = meta
        self.annotation_painter = show.AnnotationPainter()

    def targets(self, field, *, annotation_dicts):
        assert self.meta.categories is not None

        annotations = [
            AnnotationDet(self.meta.categories).set(ann['category_id'], None, ann['bbox'])
            if not ann['iscrowd']
            else AnnotationCrowd(self.meta.categories).set(ann['category_id'], ann['bbox'])
            for ann in annotation_dicts
        ]

        self._confidences(field[:, 0])
        self._regressions(field[:, 1:3], field[:, 3:5],
                          annotations=annotations)

    def predicted(self, field):
        self._confidences(field[:, 1])
        self._regressions(field[:, 2:4], field[:, 4:6],
                          annotations=self._ground_truth,
                          confidence_fields=field[:, 1],
                          uv_is_offset=False)

    def _confidences(self, confidences):
        for f in self.indices('confidence'):
            LOG.debug('%s', self.meta.categories[f])

            with self.image_canvas(self.processed_image(), margin=[0.0, 0.01, 0.05, 0.01]) as ax:
                im = ax.imshow(self.scale_scalar(confidences[f], self.meta.stride),
                               alpha=0.9, vmin=0.0, vmax=1.0, cmap=CMAP_GREENS_NAN)
                self.colorbar(ax, im)

    def _regressions(self, regression_fields, wh_fields, *,
                     annotations=None, confidence_fields=None, uv_is_offset=True):
        for f in self.indices('regression'):
            LOG.debug('%s', self.meta.categories[f])
            confidence_field = confidence_fields[f] if confidence_fields is not None else None

            with self.image_canvas(self.processed_image(), margin=[0.0, 0.01, 0.05, 0.01]) as ax:
                show.white_screen(ax, alpha=0.5)
                if annotations:
                    self.annotation_painter.annotations(ax, annotations, color='lightgray')
                q = show.quiver(ax,
                                regression_fields[f, :2],
                                confidence_field=confidence_field,
                                xy_scale=self.meta.stride, uv_is_offset=uv_is_offset,
                                cmap='Greens', clim=(0.5, 1.0), width=0.001)
                show.boxes_wh(ax, wh_fields[f, 0], wh_fields[f, 1],
                              confidence_field=confidence_field,
                              regression_field=regression_fields[f, :2],
                              xy_scale=self.meta.stride, cmap='Greens',
                              fill=False, linewidth=2,
                              regression_field_is_offset=uv_is_offset)
                if f in self.indices('margin', with_all=False):
                    show.margins(ax, regression_fields[f, :6], xy_scale=self.meta.stride)

                self.colorbar(ax, q)



import logging

import torch.utils.data

from .. import transforms


LOG = logging.getLogger(__name__)
STAT_LOG = logging.getLogger(__name__.replace('openpifpaf.', 'openpifpaf.stats.'))


class TorchDataset(torch.utils.data.Dataset):
    """Wraps a torch-based parent dataset and applies OpenPifPaf transforms."""

    def __init__(self, parent, *, preprocess=None):
        super().__init__()
        self.parent = parent
        self.preprocess = preprocess or transforms.EVAL_TRANSFORM

    def __getitem__(self, index):
        parent_data = self.parent[index]

        meta = {
            'dataset_index': index,
        }

        # preprocess image and annotations
        image, anns, meta = self.preprocess(parent_data, meta)
        LOG.debug(meta)

        return image, anns, meta

    def __len__(self):
        return len(self.parent)



from collections import defaultdict
import json
import logging
import os
import random

import PIL
import pysparkling
import torch

LOG = logging.getLogger(__name__)


class Posetrack2018(torch.utils.data.Dataset):
    """Dataset reader for Posetrack2018."""

    def __init__(self, annotation_files, data_root, *, preprocess,
                 group=None,
                 only_annotated=False,
                 max_per_sequence=None,
                 n_image_stack=1):
        super().__init__()

        if group is None:
            group = (0,)
        self.group = group

        self.preprocess = preprocess
        self.data_root = data_root
        self.only_annotated = only_annotated
        self.max_per_sequence = max_per_sequence
        self.n_image_stack = n_image_stack

        spark_context = pysparkling.Context()
        self.files = (
            spark_context
            .wholeTextFiles(annotation_files)
            .mapValues(json.loads)
            .cache()
        )
        self.annotations = self.files.flatMap(self.group_annotations).collect()

        LOG.info('sequences = %d, annotations = %d',
                 self.files.count(), len(self.annotations))

    def meta_images(self):
        return self.files.mapValues(lambda r: r['images']).collectAsMap()

    def meta_categories(self):
        return self.files.mapValues(lambda r: r['categories']).collectAsMap()

    def group_annotations(self, file_raw_annotation):
        file_name, raw_annotation = file_raw_annotation

        all_images = raw_annotation['images']
        all_annotations = raw_annotation.get('annotations', [])  # not in test set
        all_annotations_by_image_id = defaultdict(list)
        for ann in all_annotations:
            all_annotations_by_image_id[ann['image_id']].append(ann)

        frame_groups = self.group
        if not isinstance(frame_groups[0], (tuple, list)):
            frame_groups = [frame_groups]

        groups = []
        for i in range(0, len(all_images)):
            for this_group in frame_groups:
                if any(i + gi < 0 for gi in this_group):
                    continue
                image_groups = [all_images[i + gi] for gi in this_group]
                groups.append([
                    {
                        'annotation_file': file_name,
                        'image': image,
                        'annotations': all_annotations_by_image_id[image['frame_id']],
                    }
                    for image in image_groups
                ])

        if self.only_annotated:
            groups = [group for group in groups if all(s['annotations'] for s in group)]

        if self.max_per_sequence:
            if len(groups) > self.max_per_sequence:
                LOG.debug('groups per file %d -> %d', len(groups), self.max_per_sequence)
                groups = random.choices(groups, k=self.max_per_sequence)

        return groups

    def __getitem__(self, index):
        group = self.annotations[index]

        images = []
        anns = []
        metas = []

        for group_i, data in enumerate(group):
            local_file_path = os.path.join(self.data_root, data['image']['file_name'])
            with open(local_file_path, 'rb') as f:
                image = PIL.Image.open(f).convert('RGB')
                images.append(image)

            anns.append(data)

            metas.append({
                'dataset_index': index,
                'image_id': data['image']['frame_id'],
                'file_name': data['image']['file_name'],
                'local_file_path': local_file_path,
                'annotation_file': data['annotation_file'],
                'group_i': group_i,
            })

        # preprocess images and targets
        images, anns, metas = self.preprocess(images, anns, metas)
        return images, anns, metas

    def __len__(self):
        return len(self.annotations)


class Posetrack2017(torch.utils.data.Dataset):
    """Dataset reader for Posetrack."""

    def __init__(self, annotation_files, data_root, *, preprocess,
                 group=None,
                 only_annotated=False,
                 max_per_sequence=None,
                 n_image_stack=1):
        super().__init__()

        if group is None:
            group = (0,)
        self.group = group

        self.preprocess = preprocess
        self.data_root = data_root
        self.only_annotated = only_annotated
        self.max_per_sequence = max_per_sequence
        self.n_image_stack = n_image_stack

        spark_context = pysparkling.Context()
        self.files = (
            spark_context
            .wholeTextFiles(annotation_files)
            .mapValues(json.loads)
            .cache()
        )
        self.annotations = self.files.flatMap(self.group_annotations).collect()

        LOG.info('sequences = %d, annotations = %d',
                 self.files.count(), len(self.annotations))

    def meta_images(self):
        return (
            self.files
            .mapValues(lambda r: [i['image'][0]['name'] for i in r['annolist']])
            .collectAsMap()
        )

    def meta_categories(self):
        return (
            self.files
            .mapValues(lambda _: [{
                'supercategory': 'person',
                'id': 1,
                'name': 'person',
            }])
            .collectAsMap()
        )

    def group_annotations(self, file_raw_annotation):
        file_name, raw_annotation = file_raw_annotation

        all_annotations = raw_annotation.get('annolist', [])  # not in test set

        groups = []
        start_frame = 0 if not self.only_annotated else 12
        for i in range(start_frame, len(all_annotations)):
            this_group = self.group
            if isinstance(this_group, (tuple, list)) and isinstance(this_group[0], (tuple, list)):
                this_group = random.choice(this_group)

            group = [
                {
                    'annotation_file': file_name,
                    'image': {
                        'frame_id': (
                            i + gi
                            if i + gi >= 0
                            else i - 4 if i - 4 >= 0 else i
                        ),
                    },
                }
                for gi in this_group
            ]

            # extract the image info to be like 2018
            for g in group:
                annotations = all_annotations[g['image']['frame_id']]
                g['image']['file_name'] = annotations['image'][0]['name']
                # g['annotations'] = annotations
                g['annotations'] = []

            groups.append(group)

        if self.only_annotated:
            groups = [group for group in groups if all(s['annotations'] for s in group)]

        if self.max_per_sequence:
            if len(groups) > self.max_per_sequence:
                LOG.debug('groups per file %d -> %d', len(groups), self.max_per_sequence)
                groups = random.choices(groups, k=self.max_per_sequence)

        return groups

    def __getitem__(self, index):
        group = self.annotations[index]

        images = []
        anns = []
        metas = []

        for group_i, data in enumerate(group):
            with open(os.path.join(self.data_root, data['image']['file_name']), 'rb') as f:
                image = PIL.Image.open(f).convert('RGB')
                images.append(image)

            anns.append(data)

            metas.append({
                'dataset_index': index,
                'image_id': data['image']['frame_id'],
                'file_name': data['image']['file_name'],
                'annotation_file': data['annotation_file'],
                'group_i': group_i,
            })

        # preprocess images and targets
        images, anns, metas = self.preprocess(images, anns, metas)
        return images, anns, metas

    def __len__(self):
        return len(self.annotations)



import numpy as np


class Euclidean:
    """Compute Euclidean distance between a track and a new pose candidate."""

    invisible_penalty = 110.0

    def __init__(self, *, track_frames=None):
        if track_frames is None:
            track_frames = [-1]
        assert all(t < 0 for t in track_frames)

        self.valid_keypoints = None
        self.track_frames = track_frames

    def __call__(self, frame_number, pose, track, track_is_good):
        return min(
            self.distance(frame_number, pose, track, track_is_good, track_frame)
            for track_frame in self.track_frames
        )

    def distance(self, frame_number, pose, track, track_is_good, track_frame=-1):  # pylint: disable=unused-argument
        last_track_frame = track.frame_pose[-1][0]
        skipped_frames = frame_number - last_track_frame - 1
        assert skipped_frames >= 0
        if skipped_frames > 12:
            return 1000.0

        # correct track_frame with skipped_frames
        track_frame += skipped_frames
        if track_frame > -1:
            return 1000.0
        if len(track.frame_pose) < -track_frame:
            return 1000.0

        pose1 = pose.data[self.valid_keypoints]
        pose2 = track.frame_pose[track_frame][1].data[self.valid_keypoints]

        kps_distances = np.linalg.norm(pose2[:, :2] - pose1[:, :2], axis=1)
        kps_distances = np.clip(kps_distances, 0.0, self.invisible_penalty)
        kps_distances[pose1[:, 2] < 0.05] = self.invisible_penalty
        kps_distances[pose2[:, 2] < 0.05] = self.invisible_penalty
        kps_distance = np.mean(kps_distances)

        return kps_distance



import argparse

import torch

import openpifpaf

from .dataset import CocoDataset
from .constants import (
    COCO_CATEGORIES,
    COCO_KEYPOINTS,
    COCO_PERSON_SKELETON,
    COCO_PERSON_SIGMAS,
    COCO_PERSON_SCORE_WEIGHTS,
    COCO_UPRIGHT_POSE,
    DENSER_COCO_PERSON_CONNECTIONS,
    HFLIP,
)

try:
    import pycocotools.coco
    # monkey patch for Python 3 compat
    pycocotools.coco.unicode = str
except ImportError:
    pass


class CocoKp(openpifpaf.datasets.DataModule, openpifpaf.Configurable):
    _test2017_annotations = 'data-mscoco/annotations/image_info_test2017.json'
    _testdev2017_annotations = 'data-mscoco/annotations/image_info_test-dev2017.json'
    _test2017_image_dir = 'data-mscoco/images/test2017/'

    # cli configurable
    train_annotations = 'data-mscoco/annotations/person_keypoints_train2017.json'
    val_annotations = 'data-mscoco/annotations/person_keypoints_val2017.json'
    eval_annotations = val_annotations
    train_image_dir = 'data-mscoco/images/train2017/'
    val_image_dir = 'data-mscoco/images/val2017/'
    eval_image_dir = val_image_dir

    square_edge = 385
    with_dense = False
    extended_scale = False
    orientation_invariant = 0.0
    blur = 0.0
    augmentation = True
    rescale_images = 1.0
    upsample_stride = 1
    min_kp_anns = 1
    bmin = 0.1

    eval_annotation_filter = True
    eval_long_edge = 641
    eval_orientation_invariant = 0.0
    eval_extended_scale = False

    skeleton = COCO_PERSON_SKELETON

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        cif = openpifpaf.headmeta.Cif('cif', 'cocokp',
                                      keypoints=COCO_KEYPOINTS,
                                      sigmas=COCO_PERSON_SIGMAS,
                                      pose=COCO_UPRIGHT_POSE,
                                      draw_skeleton=self.skeleton,
                                      score_weights=COCO_PERSON_SCORE_WEIGHTS)
        caf = openpifpaf.headmeta.Caf('caf', 'cocokp',
                                      keypoints=COCO_KEYPOINTS,
                                      sigmas=COCO_PERSON_SIGMAS,
                                      pose=COCO_UPRIGHT_POSE,
                                      skeleton=self.skeleton)
        dcaf = openpifpaf.headmeta.Caf('caf25', 'cocokp',
                                       keypoints=COCO_KEYPOINTS,
                                       sigmas=COCO_PERSON_SIGMAS,
                                       pose=COCO_UPRIGHT_POSE,
                                       skeleton=DENSER_COCO_PERSON_CONNECTIONS,
                                       sparse_skeleton=self.skeleton,
                                       only_in_field_of_view=True)

        cif.upsample_stride = self.upsample_stride
        caf.upsample_stride = self.upsample_stride
        dcaf.upsample_stride = self.upsample_stride
        self.head_metas = [cif, caf, dcaf] if self.with_dense else [cif, caf]

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('data module CocoKp')

        group.add_argument('--cocokp-train-annotations', default=cls.train_annotations,
                           help='train annotations')
        group.add_argument('--cocokp-val-annotations', default=cls.val_annotations,
                           help='val annotations')
        group.add_argument('--cocokp-train-image-dir', default=cls.train_image_dir,
                           help='train image dir')
        group.add_argument('--cocokp-val-image-dir', default=cls.val_image_dir,
                           help='val image dir')

        group.add_argument('--cocokp-square-edge',
                           default=cls.square_edge, type=int,
                           help='square edge of input images')
        assert not cls.with_dense
        group.add_argument('--cocokp-with-dense',
                           default=False, action='store_true',
                           help='train with dense connections')
        assert not cls.extended_scale
        group.add_argument('--cocokp-extended-scale',
                           default=False, action='store_true',
                           help='augment with an extended scale range')
        group.add_argument('--cocokp-orientation-invariant',
                           default=cls.orientation_invariant, type=float,
                           help='augment with random orientations')
        group.add_argument('--cocokp-blur',
                           default=cls.blur, type=float,
                           help='augment with blur')
        assert cls.augmentation
        group.add_argument('--cocokp-no-augmentation',
                           dest='cocokp_augmentation',
                           default=True, action='store_false',
                           help='do not apply data augmentation')
        group.add_argument('--cocokp-rescale-images',
                           default=cls.rescale_images, type=float,
                           help='overall rescale factor for images')
        group.add_argument('--cocokp-upsample',
                           default=cls.upsample_stride, type=int,
                           help='head upsample stride')
        group.add_argument('--cocokp-min-kp-anns',
                           default=cls.min_kp_anns, type=int,
                           help='filter images with fewer keypoint annotations')
        group.add_argument('--cocokp-bmin',
                           default=cls.bmin, type=float,
                           help='bmin')

        # evaluation
        eval_set_group = group.add_mutually_exclusive_group()
        eval_set_group.add_argument('--cocokp-eval-test2017', default=False, action='store_true')
        eval_set_group.add_argument('--cocokp-eval-testdev2017', default=False, action='store_true')

        assert cls.eval_annotation_filter
        group.add_argument('--coco-no-eval-annotation-filter',
                           dest='coco_eval_annotation_filter',
                           default=True, action='store_false')
        group.add_argument('--coco-eval-long-edge', default=cls.eval_long_edge, type=int,
                           help='set to zero to deactivate rescaling')
        assert not cls.eval_extended_scale
        group.add_argument('--coco-eval-extended-scale', default=False, action='store_true')
        group.add_argument('--coco-eval-orientation-invariant',
                           default=cls.eval_orientation_invariant, type=float)

    @classmethod
    def configure(cls, args: argparse.Namespace):
        # extract global information
        cls.debug = args.debug
        cls.pin_memory = args.pin_memory

        # cocokp specific
        cls.train_annotations = args.cocokp_train_annotations
        cls.val_annotations = args.cocokp_val_annotations
        cls.train_image_dir = args.cocokp_train_image_dir
        cls.val_image_dir = args.cocokp_val_image_dir

        cls.square_edge = args.cocokp_square_edge
        cls.with_dense = args.cocokp_with_dense
        cls.extended_scale = args.cocokp_extended_scale
        cls.orientation_invariant = args.cocokp_orientation_invariant
        cls.blur = args.cocokp_blur
        cls.augmentation = args.cocokp_augmentation
        cls.rescale_images = args.cocokp_rescale_images
        cls.upsample_stride = args.cocokp_upsample
        cls.min_kp_anns = args.cocokp_min_kp_anns
        cls.bmin = args.cocokp_bmin

        # evaluation
        cls.eval_annotation_filter = args.coco_eval_annotation_filter
        if args.cocokp_eval_test2017:
            cls.eval_image_dir = cls._test2017_image_dir
            cls.eval_annotations = cls._test2017_annotations
            cls.annotation_filter = False
        if args.cocokp_eval_testdev2017:
            cls.eval_image_dir = cls._test2017_image_dir
            cls.eval_annotations = cls._testdev2017_annotations
            cls.annotation_filter = False
        cls.eval_long_edge = args.coco_eval_long_edge
        cls.eval_orientation_invariant = args.coco_eval_orientation_invariant
        cls.eval_extended_scale = args.coco_eval_extended_scale

        if (args.cocokp_eval_test2017 or args.cocokp_eval_testdev2017) \
                and not args.write_predictions and not args.debug:
            raise Exception('have to use --write-predictions for this dataset')

    def _preprocess(self):
        encoders = [openpifpaf.encoder.Cif(self.head_metas[0], bmin=self.bmin),
                    openpifpaf.encoder.Caf(self.head_metas[1], bmin=self.bmin)]
        if len(self.head_metas) > 2:
            encoders.append(openpifpaf.encoder.Caf(self.head_metas[2], bmin=self.bmin))

        if not self.augmentation:
            return openpifpaf.transforms.Compose([
                openpifpaf.transforms.NormalizeAnnotations(),
                openpifpaf.transforms.RescaleAbsolute(self.square_edge),
                openpifpaf.transforms.CenterPad(self.square_edge),
                openpifpaf.transforms.EVAL_TRANSFORM,
                openpifpaf.transforms.Encoders(encoders),
            ])

        if self.extended_scale:
            rescale_t = openpifpaf.transforms.RescaleRelative(
                scale_range=(0.25 * self.rescale_images,
                             2.0 * self.rescale_images),
                power_law=True, stretch_range=(0.75, 1.33))
        else:
            rescale_t = openpifpaf.transforms.RescaleRelative(
                scale_range=(0.4 * self.rescale_images,
                             2.0 * self.rescale_images),
                power_law=True, stretch_range=(0.75, 1.33))

        return openpifpaf.transforms.Compose([
            openpifpaf.transforms.NormalizeAnnotations(),
            openpifpaf.transforms.RandomApply(
                openpifpaf.transforms.HFlip(COCO_KEYPOINTS, HFLIP), 0.5),
            rescale_t,
            openpifpaf.transforms.RandomApply(
                openpifpaf.transforms.Blur(), self.blur),
            openpifpaf.transforms.RandomChoice(
                [openpifpaf.transforms.RotateBy90(),
                 openpifpaf.transforms.RotateUniform(30.0)],
                [self.orientation_invariant, 0.4],
            ),
            openpifpaf.transforms.Crop(self.square_edge, use_area_of_interest=True),
            openpifpaf.transforms.CenterPad(self.square_edge),
            openpifpaf.transforms.TRAIN_TRANSFORM,
            openpifpaf.transforms.Encoders(encoders),
        ])

    def train_loader(self):
        train_data = CocoDataset(
            image_dir=self.train_image_dir,
            ann_file=self.train_annotations,
            preprocess=self._preprocess(),
            annotation_filter=True,
            min_kp_anns=self.min_kp_anns,
            category_ids=[1],
        )
        return torch.utils.data.DataLoader(
            train_data, batch_size=self.batch_size, shuffle=not self.debug and self.augmentation,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=openpifpaf.datasets.collate_images_targets_meta)

    def val_loader(self):
        val_data = CocoDataset(
            image_dir=self.val_image_dir,
            ann_file=self.val_annotations,
            preprocess=self._preprocess(),
            annotation_filter=True,
            min_kp_anns=self.min_kp_anns,
            category_ids=[1],
        )
        return torch.utils.data.DataLoader(
            val_data, batch_size=self.batch_size, shuffle=not self.debug and self.augmentation,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=openpifpaf.datasets.collate_images_targets_meta)

    @classmethod
    def common_eval_preprocess(cls):
        rescale_t = None
        if cls.eval_extended_scale:
            assert cls.eval_long_edge
            rescale_t = [
                openpifpaf.transforms.DeterministicEqualChoice([
                    openpifpaf.transforms.RescaleAbsolute(cls.eval_long_edge),
                    openpifpaf.transforms.RescaleAbsolute((cls.eval_long_edge - 1) // 2 + 1),
                ], salt=1)
            ]
        elif cls.eval_long_edge:
            rescale_t = openpifpaf.transforms.RescaleAbsolute(cls.eval_long_edge)

        if cls.batch_size == 1:
            padding_t = openpifpaf.transforms.CenterPadTight(16)
        else:
            assert cls.eval_long_edge
            padding_t = openpifpaf.transforms.CenterPad(cls.eval_long_edge)

        orientation_t = None
        if cls.eval_orientation_invariant:
            orientation_t = openpifpaf.transforms.DeterministicEqualChoice([
                None,
                openpifpaf.transforms.RotateBy90(fixed_angle=90),
                openpifpaf.transforms.RotateBy90(fixed_angle=180),
                openpifpaf.transforms.RotateBy90(fixed_angle=270),
            ], salt=3)

        return [
            openpifpaf.transforms.NormalizeAnnotations(),
            rescale_t,
            padding_t,
            orientation_t,
        ]

    def _eval_preprocess(self):
        return openpifpaf.transforms.Compose([
            *self.common_eval_preprocess(),
            openpifpaf.transforms.ToAnnotations([
                openpifpaf.transforms.ToKpAnnotations(
                    COCO_CATEGORIES,
                    keypoints_by_category={1: self.head_metas[0].keypoints},
                    skeleton_by_category={1: self.head_metas[1].skeleton},
                ),
                openpifpaf.transforms.ToCrowdAnnotations(COCO_CATEGORIES),
            ]),
            openpifpaf.transforms.EVAL_TRANSFORM,
        ])

    def eval_loader(self):
        eval_data = CocoDataset(
            image_dir=self.eval_image_dir,
            ann_file=self.eval_annotations,
            preprocess=self._eval_preprocess(),
            annotation_filter=self.eval_annotation_filter,
            min_kp_anns=self.min_kp_anns if self.eval_annotation_filter else 0,
            category_ids=[1] if self.eval_annotation_filter else [],
        )
        return torch.utils.data.DataLoader(
            eval_data, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=False,
            collate_fn=openpifpaf.datasets.collate_images_anns_meta)

    def metrics(self):
        return [openpifpaf.metric.Coco(
            pycocotools.coco.COCO(self.eval_annotations),
            max_per_image=20,
            category_ids=[1],
            iou_type='keypoints',
        )]



"""Transform input data."""

import torchvision

from . import pair
from .annotations import AnnotationJitter, NormalizeAnnotations
from .assertion import Assert
from .compose import Compose
from .crop import Crop
from .deinterlace import Deinterlace
from .encoders import Encoders
from .hflip import HFlip
from .image import Blur, HorizontalBlur, ImageTransform, JpegCompression
from .impute import AddCrowdForIncompleteHead
from .minsize import MinSize
from .multi_scale import MultiScale
from .pad import CenterPad, CenterPadTight, SquarePad
from .preprocess import Preprocess
from .random import DeterministicEqualChoice, RandomApply, RandomChoice
from .rotate import RotateBy90, RotateUniform
from .scale import RescaleAbsolute, RescaleRelative, ScaleMix
from .toannotations import ToAnnotations, ToCrowdAnnotations, ToDetAnnotations, ToKpAnnotations
from .unclipped import UnclippedArea, UnclippedSides


EVAL_TRANSFORM = Compose([
    NormalizeAnnotations(),
    ImageTransform(torchvision.transforms.ToTensor()),
    ImageTransform(
        torchvision.transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                         std=[0.229, 0.224, 0.225]),
    ),
])


TRAIN_TRANSFORM = Compose([
    NormalizeAnnotations(),
    ImageTransform(torchvision.transforms.ColorJitter(
        brightness=0.4, contrast=0.1, saturation=0.4, hue=0.1)),
    RandomApply(JpegCompression(), 0.1),  # maybe irrelevant for COCO, but good for others
    # RandomApply(Blur(), 0.01),  # maybe irrelevant for COCO, but good for others
    ImageTransform(torchvision.transforms.RandomGrayscale(p=0.01)),
    EVAL_TRANSFORM,
])


__all__ = [
    'Preprocess',
    'AnnotationJitter', 'NormalizeAnnotations',
    'Assert',
    'Compose',
    'Crop',
    'Deinterlace',
    'Encoders',
    'HFlip',
    'Blur', 'HorizontalBlur', 'ImageTransform', 'JpegCompression',
    'AddCrowdForIncompleteHead',
    'MinSize',
    'MultiScale',
    'CenterPad', 'CenterPadTight', 'SquarePad',
    'DeterministicEqualChoice', 'RandomApply', 'RandomChoice',
    'RotateBy90', 'RotateUniform',
    'RescaleAbsolute', 'RescaleRelative', 'ScaleMix',
    'ToAnnotations', 'ToCrowdAnnotations', 'ToDetAnnotations', 'ToKpAnnotations',
    'UnclippedArea', 'UnclippedSides',
]



"""Predict poses for given images."""

import argparse
import glob
import json
import logging
import os

import torch

from . import decoder, logger, network, show, visualizer, __version__
from .predictor import Predictor

LOG = logging.getLogger(__name__)


def cli():
    parser = argparse.ArgumentParser(
        prog='python3 -m openpifpaf.predict',
        usage='%(prog)s [options] images',
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument('--version', action='version',
                        version='OpenPifPaf {version}'.format(version=__version__))

    decoder.cli(parser)
    logger.cli(parser)
    network.Factory.cli(parser)
    Predictor.cli(parser)
    show.cli(parser)
    visualizer.cli(parser)

    parser.add_argument('images', nargs='*',
                        help='input images')
    parser.add_argument('--glob',
                        help='glob expression for input images (for many images)')
    parser.add_argument('-o', '--image-output', default=None, nargs='?', const=True,
                        help='Whether to output an image, '
                             'with the option to specify the output path or directory')
    parser.add_argument('--json-output', default=None, nargs='?', const=True,
                        help='Whether to output a json file, '
                             'with the option to specify the output path or directory')
    parser.add_argument('--disable-cuda', action='store_true',
                        help='disable CUDA')
    args = parser.parse_args()

    logger.configure(args, LOG)  # logger first

    # add args.device
    args.device = torch.device('cpu')
    args.pin_memory = False
    if not args.disable_cuda and torch.cuda.is_available():
        args.device = torch.device('cuda')
        args.pin_memory = True
    LOG.info('neural network device: %s (CUDA available: %s, count: %d)',
             args.device, torch.cuda.is_available(), torch.cuda.device_count())

    decoder.configure(args)
    network.Factory.configure(args)
    Predictor.configure(args)
    show.configure(args)
    visualizer.configure(args)

    # glob
    if args.glob:
        args.images += glob.glob(args.glob)
    if not args.images:
        raise Exception("no image files given")

    return args


def out_name(arg, in_name, default_extension):
    """Determine an output name from args, input name and extension.

    arg can be:
    - none: return none (e.g. show image but don't store it)
    - True: activate this output and determine a default name
    - string:
        - not a directory: use this as the output file name
        - is a directory: use directory name and input name to form an output
    """
    if arg is None:
        return None

    if arg is True:
        return in_name + default_extension

    if os.path.isdir(arg):
        return os.path.join(
            arg,
            os.path.basename(in_name)
        ) + default_extension

    return arg


def main():
    args = cli()
    annotation_painter = show.AnnotationPainter()

    predictor = Predictor(
        visualize_image=(args.show or args.image_output is not None),
        visualize_processed_image=args.debug,
    )
    for pred, _, meta in predictor.images(args.images):
        # json output
        if args.json_output is not None:
            json_out_name = out_name(
                args.json_output, meta['file_name'], '.predictions.json')
            LOG.debug('json output = %s', json_out_name)
            with open(json_out_name, 'w') as f:
                json.dump([ann.json_data() for ann in pred], f)

        # image output
        if args.show or args.image_output is not None:
            ext = show.Canvas.out_file_extension
            image_out_name = out_name(
                args.image_output, meta['file_name'], '.predictions.' + ext)
            LOG.debug('image output = %s', image_out_name)
            image = visualizer.Base.image()
            with show.image_canvas(image, image_out_name) as ax:
                annotation_painter.annotations(ax, pred)


if __name__ == '__main__':
    main()



"""Collections of decoders: fields to annotations."""

from . import utils
from .decoder import Decoder
from .cifcaf import CifCaf
from .cifdet import CifDet
from .pose_similarity import PoseSimilarity
from .tracking_pose import TrackingPose
from .factory import cli, configure, factory

from .factory import DECODERS



import dataclasses
import logging
from typing import ClassVar

import numpy as np
import torch

from .annrescaler import AnnRescaler
from .. import headmeta
from ..visualizer import Cif as CifVisualizer
from ..utils import create_sink, mask_valid_area

LOG = logging.getLogger(__name__)


@dataclasses.dataclass
class Cif:
    meta: headmeta.Cif
    rescaler: AnnRescaler = None
    v_threshold: int = 0
    bmin: float = 0.1  #: in pixels
    visualizer: CifVisualizer = None

    side_length: ClassVar[int] = 4
    padding: ClassVar[int] = 10

    def __call__(self, image, anns, meta):
        return CifGenerator(self)(image, anns, meta)


class CifGenerator():
    def __init__(self, config: Cif):
        self.config = config

        self.rescaler = config.rescaler or AnnRescaler(
            config.meta.stride, config.meta.pose)
        self.visualizer = config.visualizer or CifVisualizer(config.meta)

        self.intensities = None
        self.fields_reg = None
        self.fields_bmin = None
        self.fields_scale = None
        self.fields_reg_l = None

        self.sink = create_sink(config.side_length)
        self.s_offset = (config.side_length - 1.0) / 2.0

    def __call__(self, image, anns, meta):
        width_height_original = image.shape[2:0:-1]

        keypoint_sets = self.rescaler.keypoint_sets(anns)
        bg_mask = self.rescaler.bg_mask(anns, width_height_original,
                                        crowd_margin=(self.config.side_length - 1) / 2)
        valid_area = self.rescaler.valid_area(meta)
        LOG.debug('valid area: %s, pif side length = %d', valid_area, self.config.side_length)

        n_fields = len(self.config.meta.keypoints)
        self.init_fields(n_fields, bg_mask)
        self.fill(keypoint_sets)
        fields = self.fields(valid_area)

        self.visualizer.processed_image(image)
        self.visualizer.targets(fields, annotation_dicts=anns)

        return fields

    def init_fields(self, n_fields, bg_mask):
        field_w = bg_mask.shape[1] + 2 * self.config.padding
        field_h = bg_mask.shape[0] + 2 * self.config.padding
        self.intensities = np.zeros((n_fields, field_h, field_w), dtype=np.float32)
        self.fields_reg = np.full((n_fields, 2, field_h, field_w), np.nan, dtype=np.float32)
        self.fields_bmin = np.full((n_fields, field_h, field_w), np.nan, dtype=np.float32)
        self.fields_scale = np.full((n_fields, field_h, field_w), np.nan, dtype=np.float32)
        self.fields_reg_l = np.full((n_fields, field_h, field_w), np.inf, dtype=np.float32)

        # bg_mask
        p = self.config.padding
        self.fields_reg_l[:, p:-p, p:-p][:, bg_mask == 0] = 1.0
        self.intensities[:, p:-p, p:-p][:, bg_mask == 0] = np.nan

    def fill(self, keypoint_sets):
        for keypoints in keypoint_sets:
            self.fill_keypoints(keypoints)

    def fill_keypoints(self, keypoints):
        scale = self.rescaler.scale(keypoints)
        for f, xyv in enumerate(keypoints):
            if xyv[2] <= self.config.v_threshold:
                continue

            joint_scale = (
                scale
                if self.config.meta.sigmas is None
                else scale * self.config.meta.sigmas[f]
            )

            self.fill_coordinate(f, xyv, joint_scale)

    def fill_coordinate(self, f, xyv, scale):
        ij = np.round(xyv[:2] - self.s_offset).astype(np.int) + self.config.padding
        minx, miny = int(ij[0]), int(ij[1])
        maxx, maxy = minx + self.config.side_length, miny + self.config.side_length
        if minx < 0 or maxx > self.intensities.shape[2] or \
           miny < 0 or maxy > self.intensities.shape[1]:
            return

        offset = xyv[:2] - (ij + self.s_offset - self.config.padding)
        offset = offset.reshape(2, 1, 1)

        # mask
        sink_reg = self.sink + offset
        sink_l = np.linalg.norm(sink_reg, axis=0)
        mask = sink_l < self.fields_reg_l[f, miny:maxy, minx:maxx]
        mask_peak = np.logical_and(mask, sink_l < 0.7)
        self.fields_reg_l[f, miny:maxy, minx:maxx][mask] = sink_l[mask]

        # update intensity
        self.intensities[f, miny:maxy, minx:maxx][mask] = 1.0
        self.intensities[f, miny:maxy, minx:maxx][mask_peak] = 1.0

        # update regression
        patch = self.fields_reg[f, :, miny:maxy, minx:maxx]
        patch[:, mask] = sink_reg[:, mask]

        # update bmin
        bmin = self.config.bmin / self.config.meta.stride
        self.fields_bmin[f, miny:maxy, minx:maxx][mask] = bmin

        # update scale
        assert np.isnan(scale) or 0.0 < scale < 100.0
        self.fields_scale[f, miny:maxy, minx:maxx][mask] = scale

    def fields(self, valid_area):
        p = self.config.padding
        intensities = self.intensities[:, p:-p, p:-p]
        fields_reg = self.fields_reg[:, :, p:-p, p:-p]
        fields_bmin = self.fields_bmin[:, p:-p, p:-p]
        fields_scale = self.fields_scale[:, p:-p, p:-p]

        mask_valid_area(intensities, valid_area)
        mask_valid_area(fields_reg[:, 0], valid_area, fill_value=np.nan)
        mask_valid_area(fields_reg[:, 1], valid_area, fill_value=np.nan)
        mask_valid_area(fields_bmin, valid_area, fill_value=np.nan)
        mask_valid_area(fields_scale, valid_area, fill_value=np.nan)

        return torch.from_numpy(np.concatenate([
            np.expand_dims(intensities, 1),
            fields_reg,
            np.expand_dims(fields_bmin, 1),
            np.expand_dims(fields_scale, 1),
        ], axis=1))



import copy
import logging
import warnings

import numpy as np
import PIL.Image
import torch

from .preprocess import Preprocess

try:
    import cv2
except ImportError:
    cv2 = None

try:
    import scipy.ndimage
except ImportError:
    scipy = None  # pylint: disable=invalid-name

LOG = logging.getLogger(__name__)


def _scale(image, anns, meta, target_w, target_h, resample, *, fast=False):
    """target_w and target_h as integers

    Internally, resample in Pillow are aliases:
    PIL.Image.Resampling.BILINEAR = 2
    PIL.Image.Resampling.BICUBIC = 3
    """
    meta = copy.deepcopy(meta)
    anns = copy.deepcopy(anns)
    w, h = image.size

    assert resample in (0, 2, 3)

    # scale image
    if fast and cv2 is not None:
        LOG.debug('using OpenCV for fast rescale')
        if resample == 0:
            cv_interpoltation = cv2.INTER_NEAREST
        elif resample == 2:
            cv_interpoltation = cv2.INTER_LINEAR
        elif resample == 3:
            cv_interpoltation = cv2.INTER_CUBIC
        else:
            raise NotImplementedError('resample of {} not implemented for OpenCV'.format(resample))
        im_np = np.asarray(image)
        im_np = cv2.resize(im_np, (target_w, target_h), interpolation=cv_interpoltation)
        image = PIL.Image.fromarray(im_np)
    elif fast:
        LOG.debug('Requested fast resizing without OpenCV. Using Pillow. '
                  'Install OpenCV for even faster image resizing.')
        image = image.resize((target_w, target_h), resample)
    else:
        order = resample
        if order == 2:
            order = 1

        im_np = np.asarray(image)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            im_np = scipy.ndimage.zoom(im_np, (target_h / h, target_w / w, 1), order=order)
        image = PIL.Image.fromarray(im_np)

    LOG.debug('before resize = (%f, %f), after = %s', w, h, image.size)
    assert image.size[0] == target_w
    assert image.size[1] == target_h

    # rescale keypoints
    x_scale = (image.size[0] - 1) / (w - 1)
    y_scale = (image.size[1] - 1) / (h - 1)
    scale_factors = np.array((x_scale, y_scale))
    for ann in anns:
        ann['keypoints'][:, [0, 1]] *= np.expand_dims(scale_factors, 0)
        ann['bbox'][:2] *= scale_factors
        ann['bbox'][2:] *= scale_factors

    # adjust meta
    LOG.debug('meta before: %s', meta)
    meta['offset'] *= scale_factors
    meta['scale'] *= scale_factors
    meta['valid_area'][:2] *= scale_factors
    meta['valid_area'][2:] *= scale_factors
    LOG.debug('meta after: %s', meta)

    return image, anns, meta


class RescaleRelative(Preprocess):
    """Rescale relative to input image."""

    def __init__(self, scale_range=(0.5, 1.0), *,
                 resample=PIL.Image.Resampling.BILINEAR,
                 absolute_reference=None,
                 fast=False,
                 power_law=False,
                 stretch_range=None):
        self.scale_range = scale_range
        self.resample = resample
        self.absolute_reference = absolute_reference
        self.fast = fast
        self.power_law = power_law
        self.stretch_range = stretch_range

    def __call__(self, image, anns, meta):
        if isinstance(self.scale_range, tuple):
            if self.power_law:
                rnd_range = np.log2(self.scale_range[0]), np.log2(self.scale_range[1])
                log2_scale_factor = (
                    rnd_range[0]
                    + torch.rand(1).item() * (rnd_range[1] - rnd_range[0])
                )

                scale_factor = 2 ** log2_scale_factor
                LOG.debug('rnd range = %s, log2_scale_Factor = %f, scale factor = %f',
                          rnd_range, log2_scale_factor, scale_factor)
            else:
                scale_factor = (
                    self.scale_range[0]
                    + torch.rand(1).item() * (self.scale_range[1] - self.scale_range[0])
                )
        else:
            scale_factor = self.scale_range

        w, h = image.size
        if self.absolute_reference is not None:
            if w > h:
                h *= self.absolute_reference / w
                w = self.absolute_reference
            else:
                w *= self.absolute_reference / h
                h = self.absolute_reference

        stretch_factor = 1.0
        if self.stretch_range is not None:
            stretch_factor = (
                self.stretch_range[0]
                + torch.rand(1).item() * (self.stretch_range[1] - self.stretch_range[0])
            )

        target_w, target_h = int(w * scale_factor * stretch_factor), int(h * scale_factor)
        return _scale(image, anns, meta, target_w, target_h, self.resample, fast=self.fast)


class RescaleAbsolute(Preprocess):
    """Rescale to a given size."""

    def __init__(self, long_edge, *, fast=False, resample=PIL.Image.Resampling.BILINEAR):
        self.long_edge = long_edge
        self.fast = fast
        self.resample = resample

    def __call__(self, image, anns, meta):
        w, h = image.size

        this_long_edge = self.long_edge
        if isinstance(this_long_edge, (tuple, list)):
            this_long_edge = torch.randint(
                int(this_long_edge[0]),
                int(this_long_edge[1]), (1,)
            ).item()

        s = this_long_edge / max(h, w)
        if h > w:
            target_w, target_h = int(w * s), int(this_long_edge)
        else:
            target_w, target_h = int(this_long_edge), int(h * s)
        return _scale(image, anns, meta, target_w, target_h, self.resample, fast=self.fast)


class ScaleMix(Preprocess):
    def __init__(self, scale_threshold, *,
                 upscale_factor=2.0,
                 downscale_factor=0.5,
                 resample=PIL.Image.Resampling.BILINEAR):
        self.scale_threshold = scale_threshold
        self.upscale_factor = upscale_factor
        self.downscale_factor = downscale_factor
        self.resample = resample

    def __call__(self, image, anns, meta):
        scales = np.array([
            np.sqrt(ann['bbox'][2] * ann['bbox'][3])
            for ann in anns if (not getattr(ann, 'iscrowd', False)
                                and np.any(ann['keypoints'][:, 2] > 0.0))
        ])
        LOG.debug('scale threshold = %f, scales = %s', self.scale_threshold, scales)
        if not scales.shape[0]:
            return image, anns, meta

        all_above_threshold = np.all(scales > self.scale_threshold)
        all_below_threshold = np.all(scales < self.scale_threshold)
        if not all_above_threshold and \
           not all_below_threshold:
            return image, anns, meta

        w, h = image.size
        if all_above_threshold:
            target_w, target_h = int(w / 2), int(h / 2)
        else:
            target_w, target_h = int(w * 2), int(h * 2)
        LOG.debug('scale mix from (%d, %d) to (%d, %d)', w, h, target_w, target_h)
        return _scale(image, anns, meta, target_w, target_h, self.resample)



import argparse
import logging
import socket
import sys

import torch

LOG = logging.getLogger(__name__)


def cli(parser: argparse.ArgumentParser):
    group = parser.add_argument_group('logger')
    group.add_argument('-q', '--quiet', default=False, action='store_true',
                       help='only show warning messages or above')
    group.add_argument('--debug', default=False, action='store_true',
                       help='print debug messages')
    group.add_argument('--log-stats', default=False, action='store_true',
                       help='enable stats logging')


def configure(args: argparse.Namespace, local_logger=None):
    log_level = logging.INFO
    if args.quiet:
        log_level = logging.WARNING
    if args.debug:
        assert not args.quiet
        log_level = logging.DEBUG

    stdout_handler = logging.StreamHandler(sys.stdout)
    if args.log_stats:
        # pylint: disable=import-outside-toplevel
        from pythonjsonlogger import jsonlogger
        stdout_handler.setFormatter(
            jsonlogger.JsonFormatter('(message) (levelname) (name)'))
        logging.basicConfig(handlers=[stdout_handler])
        logging.getLogger('openpifpaf.stats').setLevel(logging.DEBUG)
    else:
        logging.basicConfig(handlers=[stdout_handler])

    # set log level for openpifpaf and all openpifpaf plugins
    for logger_name in logging.root.manager.loggerDict:  # pylint: disable=no-member
        if '.' in logger_name or not logger_name.startswith('openpifpaf'):
            continue
        logging.getLogger(logger_name).setLevel(log_level)

    if local_logger is not None:
        local_logger.setLevel(log_level)

    if log_level in (logging.WARNING, logging.ERROR):
        torch.ops.openpifpaf.set_quiet(True)


def train_configure(args):
    if torch.distributed.is_initialized():
        rank_prefix = 'Rank {}/{}'.format(
            torch.distributed.get_rank(), torch.distributed.get_world_size())
        formatter = logging.Formatter(rank_prefix + ' - %(levelname)s:%(name)s:%(message)s')
        for handler in logging.getLogger().handlers:
            handler.setFormatter(formatter)

    if torch.distributed.is_initialized() and torch.distributed.get_rank() != 0:
        return

    # pylint: disable=import-outside-toplevel,cyclic-import
    from pythonjsonlogger import jsonlogger
    from . import __version__
    from .plugin import versions as plugin_versions

    file_handler = logging.FileHandler(args.output + '.log', mode='w')
    file_handler.setFormatter(
        jsonlogger.JsonFormatter('%(message) %(levelname) %(name) %(asctime)'))
    logging.getLogger('openpifpaf').addHandler(file_handler)

    LOG.info({
        'type': 'process',
        'argv': sys.argv,
        'args': vars(args),
        'version': __version__,
        'plugin_versions': plugin_versions(),
        'hostname': socket.gethostname(),
    })



class Base:
    def accumulate(self, predictions, image_meta, *, ground_truth=None):
        """For every image, accumulate that image's predictions into this metric.

        :param predictions: List of predictions for one image.
        :param image_meta: Meta dictionary for this image as returned by the data loader.
        :param ground_truth: Ground truth information as produced by the eval
            loader. Optional because some metrics (i.e. pycocotools) read
            ground truth separately.
        """
        raise NotImplementedError

    def stats(self):
        """Return a dictionary of summary statistics.

        The dictionary should be of the following form and can contain
        an arbitrary number of entries with corresponding labels:

        .. code-block::

            {
                'stats': [0.1234, 0.5134],
                'text_labels': ['AP', 'AP0.50'],
            }
        """
        raise NotImplementedError

    def write_predictions(self, filename, *, additional_data=None):
        """Write predictions to a file.

        This is used to produce a metric-compatible output of predictions.
        It is used for test challenge submissions where a remote server
        holds the private test set.

        :param filename: Output filename of prediction file.
        :param additional_data: Additional information that might be worth saving
            along with the predictions.
        """
        raise NotImplementedError



import glob
import os
import setuptools
import sys
import torch.utils.cpp_extension


# This is needed for versioneer to be importable when building with PEP 517.
# See <https://github.com/warner/python-versioneer/issues/193> and links
# therein for more information.
sys.path.append(os.path.dirname(__file__))
import versioneer


EXTENSIONS = []
CMD_CLASS = {}


def add_cpp_extension():
    extra_compile_args = [
        '-std=c++17' if not sys.platform.startswith('win') else '/std:c++17',
    ]
    extra_link_args = []
    define_macros = [
        ('_SILENCE_ALL_CXX17_DEPRECATION_WARNINGS', None),  # mostly for the pytorch codebase
    ]

    if sys.platform.startswith('win'):
        extra_compile_args += ['/permissive']
        define_macros += [('OPENPIFPAF_DLLEXPORT', None)]

    if os.getenv('DEBUG', '0') == '1':
        print('DEBUG mode')
        if sys.platform.startswith('linux'):
            extra_compile_args += ['-g', '-Og']
            extra_compile_args += [
                '-Wuninitialized',
                # '-Werror',  # fails in pytorch code, but would be nice to have in CI
            ]
        define_macros += [('DEBUG', None)]

    this_dir = os.path.dirname(os.path.abspath(__file__))
    csrc_dir = os.path.join('src', 'openpifpaf', 'csrc')
    EXTENSIONS.append(
        torch.utils.cpp_extension.CppExtension(
            'openpifpaf._cpp',
            glob.glob(os.path.join(csrc_dir, 'src', '**', '*.cpp'), recursive=True),
            depends=glob.glob(os.path.join(this_dir, csrc_dir, 'include', '**', '*.hpp'), recursive=True),
            include_dirs=[os.path.join(this_dir, csrc_dir, 'include')],
            define_macros=define_macros,
            extra_compile_args=extra_compile_args,
            extra_link_args=extra_link_args,
        )
    )
    assert 'build_ext' not in CMD_CLASS, f"build_ext in CMD_CLASS: {CMD_CLASS}"
    CMD_CLASS['build_ext'] = torch.utils.cpp_extension.BuildExtension.with_options(no_python_abi_suffix=True)


add_cpp_extension()
CMD_CLASS = versioneer.get_cmdclass(CMD_CLASS)
setuptools.setup(
    name='openpifpaf',
    version=versioneer.get_version(),
    license='GNU AGPLv3',
    description='PifPaf: Composite Fields for Human Pose Estimation',
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    author='Sven Kreiss',
    author_email='research@svenkreiss.com',
    url='https://github.com/openpifpaf/openpifpaf',

    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},
    package_data={
        'openpifpaf': ['*.dll', '*.dylib', '*.so'],
    },
    cmdclass=CMD_CLASS,
    ext_modules=EXTENSIONS,
    zip_safe=False,

    python_requires='>=3.7',
    install_requires=[
        'importlib_metadata!=3.8.0',  # temporary for pytest
        'numpy>=1.16',
        'pysparkling',  # for log analysis
        'python-json-logger',
        'torch==1.11.0',
        'torchvision==0.12.0',
        'pillow!=8.3.0',  # exclusion torchvision 0.10.0 compatibility
        'dataclasses; python_version<"3.7"',
    ],
    extras_require={
        'backbones': [
            'timm>=0.4.9',  # For Swin Transformer and XCiT
            'einops>=0.3',  # required for BotNet
        ],
        'dev': [
            'flameprof',
            'jupyter-book>=0.9.1',
            'matplotlib>=3.3',
            'nbdime',
            'nbstripout',
            'scipy',
            'sphinx-book-theme',
            'wheel',
        ],
        'onnx': [
            'onnx',
            'onnxruntime',
            'onnx-simplifier>=0.2.9; python_version<"3.9"',  # Python 3.9 not supported yet
            'protobuf<4',  # temporary explicit dependency until tests pass again
        ],
        'coreml': [
            'coremltools>=5.0b3',
        ],
        'test': [
            'cpplint',
            'nbconvert',
            'nbstripout',
            'nbval',
            'opencv-python',
            'pylint<2.9.4',  # avoid 2.9.4 and up for time.perf_counter deprecation warnings
            'pycodestyle',
            'pytest',
            'thop; python_version<"3.10"',
        ],
        'train': [
            'matplotlib>=3.3',  # required by pycocotools
            'pycocotools>=2.0.1',  # pre-install cython (currently incompatible with numpy 1.18 or above)
            'scipy',
            'xtcocotools>=1.5; sys_platform == "linux"',  # required for wholebody eval, only wheels and only for linux on pypi
        ],
    },
)



from typing import List

from .preprocess import Preprocess


class Compose(Preprocess):
    """Execute given transforms in sequential order."""
    def __init__(self, preprocess_list: List[Preprocess]):
        self.preprocess_list = preprocess_list

    def __call__(self, *args):
        for p in self.preprocess_list:
            if p is None:
                continue
            args = p(*args)

        return args



"""Convert a set of keypoint coordinates into target fields.

Takes an annotation from a dataset and turns it into the
ground truth for a field.
"""

from .annrescaler import AnnRescaler, AnnRescalerDet, TrackingAnnRescaler
from .factory import cli, configure
from .caf import Caf
from .cif import Cif
from .cifdet import CifDet
from .single_image import SingleImage
from .tcaf import Tcaf



import logging

import torch

LOG = logging.getLogger(__name__)


class MultiHeadLoss(torch.nn.Module):
    task_sparsity_weight = 0.0

    def __init__(self, losses, lambdas):
        super().__init__()

        if not lambdas:
            lambdas = [1.0 for l in losses for _ in l.field_names]
        assert all(lam >= 0.0 for lam in lambdas)

        self.losses = torch.nn.ModuleList(losses)
        self.lambdas = lambdas

        self.field_names = [n for l in self.losses for n in l.field_names]
        assert len(self.field_names) == len(self.lambdas)
        LOG.info('multihead loss: %s, %s', self.field_names, self.lambdas)

    def forward(self, head_fields, head_targets):  # pylint: disable=arguments-differ
        assert len(self.losses) == len(head_fields)
        assert len(self.losses) <= len(head_targets)
        assert self.task_sparsity_weight == 0.0  # TODO implement
        flat_head_losses = [ll
                            for l, f, t in zip(self.losses, head_fields, head_targets)
                            for ll in l(f, t)]

        assert len(self.lambdas) == len(flat_head_losses)
        loss_values = [lam * l
                       for lam, l in zip(self.lambdas, flat_head_losses)
                       if l is not None]
        total_loss = sum(loss_values) if loss_values else None

        return total_loss, flat_head_losses


class MultiHeadLossAutoTuneKendall(torch.nn.Module):
    task_sparsity_weight = 0.0

    def __init__(self, losses, lambdas, *, sparse_task_parameters=None, tune=None):
        """Auto-tuning multi-head loss.

        Uses idea from "Multi-Task Learning Using Uncertainty to Weigh Losses
        for Scene Geometry and Semantics" by Kendall, Gal and Cipolla.

        Individual losses must not be negative for Kendall's prescription.

        In the common setting, use lambdas of zero and one to deactivate and
        activate the tasks you want to train. Less common, if you have
        secondary tasks, you can reduce their importance by choosing a
        lambda value between zero and one.
        """
        super().__init__()

        if not lambdas:
            lambdas = [1.0 for l in losses for _ in l.field_names]
        assert all(lam >= 0.0 for lam in lambdas)

        self.losses = torch.nn.ModuleList(losses)
        self.lambdas = lambdas
        self.sparse_task_parameters = sparse_task_parameters
        self.tune = tune

        self.log_sigmas = torch.nn.Parameter(
            torch.zeros((len(lambdas),), dtype=torch.float64),
            requires_grad=True,
        )

        self.field_names = [n for l in self.losses for n in l.field_names]
        LOG.info('multihead loss with autotune: %s', self.field_names)
        assert len(self.field_names) == len(self.lambdas)
        assert len(self.field_names) == len(self.log_sigmas)

        if self.tune is None:
            def tune_from_name(name):
                if '.vec' in name:
                    return 'none'
                if '.scale' in name:
                    return 'laplace'
                return 'gauss'
            self.tune = [
                tune_from_name(n)
                for l in self.losses for n in l.field_names
            ]
        LOG.info('tune config: %s', self.tune)

    def batch_meta(self):
        return {'mtl_sigmas': [round(float(s), 3) for s in self.log_sigmas.exp()]}

    def forward(self, head_fields, head_targets):
        LOG.debug('losses = %d, fields = %d, targets = %d',
                  len(self.losses), len(head_fields), len(head_targets))
        assert len(self.losses) == len(head_fields)
        assert len(self.losses) <= len(head_targets)
        flat_head_losses = [ll
                            for l, f, t in zip(self.losses, head_fields, head_targets)
                            for ll in l(f, t)]

        assert len(self.lambdas) == len(flat_head_losses)
        assert len(self.log_sigmas) == len(flat_head_losses)
        constrained_log_sigmas = 3.0 * torch.tanh(self.log_sigmas / 3.0)

        def tuned_loss(tune, log_sigma, loss):
            if tune == 'none':
                return loss
            if tune == 'laplace':
                # this is the negative ln of a Laplace with
                # ln(2) = 0.694
                return 0.694 + log_sigma + loss * torch.exp(-log_sigma)
            if tune == 'gauss':
                # this is the negative ln of a Gaussian with
                # ln(sqrt(2pi)) = 0.919
                return 0.919 + log_sigma + loss * 0.5 * torch.exp(-2.0 * log_sigma)
            raise Exception('unknown tune: {}'.format(tune))

        loss_values = [
            lam * tuned_loss(t, log_sigma, l)
            for lam, t, log_sigma, l in zip(
                self.lambdas, self.tune, constrained_log_sigmas, flat_head_losses)
            if l is not None
        ]
        total_loss = sum(loss_values) if loss_values else None

        if self.task_sparsity_weight and self.sparse_task_parameters is not None:
            head_sparsity_loss = sum(
                # torch.norm(param, p=1)
                param.abs().max(dim=1)[0].clamp(min=1e-6).sum()
                for param in self.sparse_task_parameters
            )
            LOG.debug('l1 head sparsity loss = %f (total = %f)', head_sparsity_loss, total_loss)
            total_loss = total_loss + self.task_sparsity_weight * head_sparsity_loss

        return total_loss, flat_head_losses


class MultiHeadLossAutoTuneVariance(torch.nn.Module):
    task_sparsity_weight = 0.0

    def __init__(self, losses, lambdas, *, sparse_task_parameters=None):
        """Auto-tuning multi-head loss based on loss-variance.

        In the common setting, use lambdas of zero and one to deactivate and
        activate the tasks you want to train. Less common, if you have
        secondary tasks, you can reduce their importance by choosing a
        lambda value between zero and one.
        """
        super().__init__()

        if not lambdas:
            lambdas = [1.0 for l in losses for _ in l.field_names]
        assert all(lam >= 0.0 for lam in lambdas)

        self.losses = torch.nn.ModuleList(losses)
        self.lambdas = lambdas
        self.sparse_task_parameters = sparse_task_parameters

        self.epsilons = torch.ones((len(lambdas),), dtype=torch.float64)
        # choose a prime number for the buffer length:
        # for multiple tasks, prevents that always the same buffer_index is
        # skipped which would mean that some nan values will remain forever
        self.buffer = torch.full((len(lambdas), 53), float('nan'), dtype=torch.float64)
        self.buffer_index = -1

        self.field_names = [n for l in self.losses for n in l.field_names]
        LOG.info('multihead loss with autotune: %s', self.field_names)
        assert len(self.field_names) == len(self.lambdas)
        assert len(self.field_names) == len(self.epsilons)

    def batch_meta(self):
        return {'mtl_sigmas': [round(float(s), 3) for s in self.epsilons]}

    def forward(self, head_fields, head_targets):
        LOG.debug('losses = %d, fields = %d, targets = %d',
                  len(self.losses), len(head_fields), len(head_targets))
        assert len(self.losses) == len(head_fields)
        assert len(self.losses) <= len(head_targets)
        flat_head_losses = [ll
                            for l, f, t in zip(self.losses, head_fields, head_targets)
                            for ll in l(f, t)]

        self.buffer_index = (self.buffer_index + 1) % self.buffer.shape[1]
        for i, ll in enumerate(flat_head_losses):
            if not hasattr(ll, 'data'):  # e.g. when ll is None
                continue
            self.buffer[i, self.buffer_index] = ll.data
        self.epsilons = torch.sqrt(
            torch.mean(self.buffer**2, dim=1)
            - torch.sum(self.buffer, dim=1)**2 / self.buffer.shape[1]**2
        )
        self.epsilons[torch.isnan(self.epsilons)] = 10.0
        self.epsilons = self.epsilons.clamp(0.01, 100.0)

        # normalize sum 1/eps to the number of sub-losses
        LOG.debug('eps before norm: %s', self.epsilons)
        self.epsilons = self.epsilons * torch.sum(1.0 / self.epsilons) / self.epsilons.shape[0]
        LOG.debug('eps after norm: %s', self.epsilons)

        assert len(self.lambdas) == len(flat_head_losses)
        assert len(self.epsilons) == len(flat_head_losses)
        loss_values = [
            lam * l / eps
            for lam, eps, l in zip(self.lambdas, self.epsilons, flat_head_losses)
            if l is not None
        ]
        total_loss = sum(loss_values) if loss_values else None

        if self.task_sparsity_weight and self.sparse_task_parameters is not None:
            head_sparsity_loss = sum(
                # torch.norm(param, p=1)
                param.abs().max(dim=1)[0].clamp(min=1e-6).sum()
                for param in self.sparse_task_parameters
            )
            LOG.debug('l1 head sparsity loss = %f (total = %f)', head_sparsity_loss, total_loss)
            total_loss = total_loss + self.task_sparsity_weight * head_sparsity_loss

        return total_loss, flat_head_losses



import copy
import logging

import numpy as np
import torch

from .preprocess import Preprocess

LOG = logging.getLogger(__name__)


class Crop(Preprocess):
    """Random cropping."""

    def __init__(self, long_edge, use_area_of_interest=True):
        self.long_edge = long_edge
        self.use_area_of_interest = use_area_of_interest

    def __call__(self, image, anns, meta):
        meta = copy.deepcopy(meta)
        anns = copy.deepcopy(anns)
        original_valid_area = meta['valid_area'].copy()

        image, anns, ltrb = self.crop(image, anns, meta['valid_area'])
        meta['offset'] += ltrb[:2]

        new_wh = image.size
        LOG.debug('valid area before crop of %s: %s', ltrb, original_valid_area)
        # process crops from left and top
        meta['valid_area'][:2] = np.maximum(0.0, original_valid_area[:2] - ltrb[:2])
        # process crops from right and bottom
        new_rb_corner = original_valid_area[:2] + original_valid_area[2:] - ltrb[:2]
        new_rb_corner = np.maximum(meta['valid_area'][:2], new_rb_corner)
        new_rb_corner = np.minimum(new_wh, new_rb_corner)
        meta['valid_area'][2:] = new_rb_corner - meta['valid_area'][:2]
        LOG.debug('valid area after crop: %s', meta['valid_area'])

        # clip bounding boxes
        for ann in anns:
            unclipped_bbox = ann['bbox'].copy()
            ann['bbox'][:2] = np.maximum(meta['valid_area'][:2], ann['bbox'][:2])
            new_rb = unclipped_bbox[:2] + unclipped_bbox[2:]
            new_rb = np.maximum(ann['bbox'][:2], new_rb)
            new_rb = np.minimum(meta['valid_area'][:2] + meta['valid_area'][2:], new_rb)
            ann['bbox'][2:] = new_rb - ann['bbox'][:2]
        anns = [ann for ann in anns if ann['bbox'][2] > 0.0 and ann['bbox'][3] > 0.0]

        return image, anns, meta

    @staticmethod
    def area_of_interest(anns, valid_area):
        """area that contains annotations with keypoints"""

        points_of_interest = [
            xy
            for ann in anns
            if not ann.get('iscrowd', False)
            for xy in [ann['bbox'][:2], ann['bbox'][:2] + ann['bbox'][2:]]
        ]
        if not points_of_interest:
            return valid_area
        points_of_interest = np.stack(points_of_interest, axis=0)
        min_xy = np.min(points_of_interest, axis=0) - 50
        max_xy = np.max(points_of_interest, axis=0) + 50

        # Make sure to stay inside of valid area.
        left = np.clip(min_xy[0], valid_area[0], valid_area[0] + valid_area[2] - 1)
        top = np.clip(min_xy[1], valid_area[1], valid_area[1] + valid_area[3] - 1)
        right = np.clip(max_xy[0], left + 1, valid_area[0] + valid_area[2])
        bottom = np.clip(max_xy[1], top + 1, valid_area[1] + valid_area[3])

        return (left, top, right - left, bottom - top)

    @staticmethod
    def random_location_1d(image_length,
                           valid_min, valid_length,
                           interest_min, interest_length,
                           crop_length,
                           tail=0.1, shift=0.0, fix_inconsistent=True):
        if image_length <= crop_length:
            return 0

        if fix_inconsistent:
            # relevant for tracking with inconsistent image sizes
            # (e.g. with RandomizeOneFrame augmentation)
            valid_min = np.clip(valid_min, 0, image_length)
            valid_length = np.clip(valid_length, 0, image_length - valid_min)
            interest_min = np.clip(interest_min, 0, image_length)
            interest_length = np.clip(interest_length, 0, image_length - interest_min)

        sticky_rnd = -tail + 2 * tail * torch.rand((1,)).item()
        sticky_rnd = np.clip(sticky_rnd, 0.0, 1.0)

        if interest_length > crop_length:
            # crop within area of interest
            sticky_rnd = np.clip(sticky_rnd + shift / interest_length, 0.0, 1.0)
            offset = interest_min + (interest_length - crop_length) * sticky_rnd
            return int(offset)

        # from above: interest_length < crop_length
        min_v = interest_min + interest_length - crop_length
        max_v = interest_min

        if valid_length > crop_length:
            # clip to valid area
            min_v = max(min_v, valid_min)
            max_v = max(min_v, min(max_v, valid_min + valid_length - crop_length))
        elif image_length > crop_length:
            # clip to image
            min_v = max(min_v, 0)
            max_v = max(min_v, min(max_v, 0 + image_length - crop_length))

        # image constraint
        min_v = np.clip(min_v, 0, image_length - crop_length)
        max_v = np.clip(max_v, 0, image_length - crop_length)

        assert max_v >= min_v
        sticky_rnd = np.clip(sticky_rnd + shift / (max_v - min_v + 1e-3), 0.0, 1.0)
        offset = min_v + (max_v - min_v) * sticky_rnd
        return int(offset)

    def crop(self, image, anns, valid_area):
        if self.use_area_of_interest:
            area_of_interest = self.area_of_interest(anns, valid_area)
        else:
            area_of_interest = valid_area

        w, h = image.size
        x_offset, y_offset = 0, 0
        if w > self.long_edge:
            x_offset = self.random_location_1d(
                w - 1,
                valid_area[0], valid_area[2],
                area_of_interest[0], area_of_interest[2],
                self.long_edge,
            )
        if h > self.long_edge:
            y_offset = self.random_location_1d(
                h - 1,
                valid_area[1], valid_area[3],
                area_of_interest[1], area_of_interest[3],
                self.long_edge
            )
        LOG.debug('crop offsets (%d, %d)', x_offset, y_offset)

        # crop image
        new_w = min(self.long_edge, w - x_offset)
        new_h = min(self.long_edge, h - y_offset)
        # ltrb might be confusing name:
        # it's the coordinates of the top-left corner and the coordinates
        # of the bottom right corner
        ltrb = (x_offset, y_offset, x_offset + new_w, y_offset + new_h)
        image = image.crop(ltrb)

        # crop keypoints
        for ann in anns:
            ann['keypoints'][:, 0] -= x_offset
            ann['keypoints'][:, 1] -= y_offset
            ann['bbox'][0] -= x_offset
            ann['bbox'][1] -= y_offset

        return image, anns, np.array(ltrb)



from .painters import KeypointPainter, CrowdPainter, DetectionPainter

PAINTERS = {
    'Annotation': KeypointPainter,
    'AnnotationCrowd': CrowdPainter,
    'AnnotationDet': DetectionPainter,
}


class AnnotationPainter:
    def __init__(self, *,
                 xy_scale=1.0,
                 painters=None):
        self.painters = {annotation_type: painter(xy_scale=xy_scale)
                         for annotation_type, painter in PAINTERS.items()}

        if painters:
            for annotation_type, painter in painters.items():
                self.painters[annotation_type] = painter

    def annotations(self, ax, annotations, *,
                    color=None, colors=None, texts=None, subtexts=None, **kwargs):
        for i, ann in enumerate(annotations):
            if colors is not None:
                this_color = colors[i]
            elif color is not None:
                this_color = color
            elif getattr(ann, 'id_', None):
                this_color = ann.id_
            else:
                this_color = i

            text = None
            if texts is not None:
                text = texts[i]

            subtext = None
            if subtexts is not None:
                subtext = subtexts[i]

            painter = self.painters[ann.__class__.__name__]
            painter.annotation(ax, ann, color=this_color, text=text, subtext=subtext, **kwargs)



# Copyright (c) by Phil Wang licensed under the MIT license. [see docs/LICENSE.BOTNET for details]

import torch
from torch import nn, einsum

try:
    from einops import rearrange
except ImportError:
    rearrange = None


def pair(x):
    return (x, x) if not isinstance(x, tuple) else x


def expand_dim(t, dim, k_num):
    t = t.unsqueeze(dim=dim)
    expand_shape = [-1] * len(t.shape)
    expand_shape[dim] = k_num
    return t.expand(*expand_shape)


def rel_to_abs(x):
    b, h, l, _, device, dtype = *x.shape, x.device, x.dtype
    dd = {'device': device, 'dtype': dtype}
    col_pad = torch.zeros((b, h, l, 1), **dd)
    x = torch.cat((x, col_pad), dim=3)
    flat_x = rearrange(x, 'b h l c -> b h (l c)')
    flat_pad = torch.zeros((b, h, l - 1), **dd)
    flat_x_padded = torch.cat((flat_x, flat_pad), dim=2)
    final_x = flat_x_padded.reshape(b, h, l + 1, 2 * l - 1)
    final_x = final_x[:, :, :l, (l - 1):]
    return final_x


def relative_logits_1d(q_num, rel_k):
    b, heads, h, w, _ = q_num.shape
    logits = einsum('b h x y d, r d -> b h x y r', q_num, rel_k)
    logits = rearrange(logits, 'b h x y r -> b (h x) y r')
    logits = rel_to_abs(logits)
    logits = logits.reshape(b, heads, h, w, w)
    logits = expand_dim(logits, dim=3, k_num=h)
    return logits

# positional embeddings


class AbsPosEmb(nn.Module):
    def __init__(
        self,
        fmap_size,
        dim_head
    ):
        super().__init__()
        height, width = pair(fmap_size)
        scale = dim_head ** -0.5
        self.height = nn.Parameter(torch.randn(height, dim_head) * scale)
        self.width = nn.Parameter(torch.randn(width, dim_head) * scale)

    def forward(self, q_num):
        emb = rearrange(self.height, 'h d -> h () d') + rearrange(self.width, 'w d -> () w d')
        emb = rearrange(emb, ' h w d -> (h w) d')
        logits = einsum('b h i d, j d -> b h i j', q_num, emb)
        return logits


class RelPosEmb(nn.Module):
    def __init__(
        self,
        fmap_size,
        dim_head
    ):
        super().__init__()
        height, width = pair(fmap_size)
        scale = dim_head ** -0.5
        self.fmap_size = fmap_size
        self.rel_height = nn.Parameter(torch.randn(height * 2 - 1, dim_head) * scale)
        self.rel_width = nn.Parameter(torch.randn(width * 2 - 1, dim_head) * scale)

    def forward(self, q_num):
        h, w = self.fmap_size

        q_num = rearrange(q_num, 'b h (x y) d -> b h x y d', x=h, y=w)
        rel_logits_w = relative_logits_1d(q_num, self.rel_width)
        rel_logits_w = rearrange(rel_logits_w, 'b h x i y j-> b h (x y) (i j)')

        q_num = rearrange(q_num, 'b h x y d -> b h y x d')
        rel_logits_h = relative_logits_1d(q_num, self.rel_height)
        rel_logits_h = rearrange(rel_logits_h, 'b h x i y j -> b h (y x) (j i)')
        return rel_logits_w + rel_logits_h

# classes


class Attention(nn.Module):
    def __init__(
        self,
        *,
        dim,
        fmap_size,
        heads=4,
        dim_head=128,
        rel_pos_emb=False
    ):
        super().__init__()
        self.heads = heads
        self.scale = dim_head ** -0.5
        inner_dim = heads * dim_head

        self.to_qkv = nn.Conv2d(dim, inner_dim * 3, 1, bias=False)

        rel_pos_class = AbsPosEmb if not rel_pos_emb else RelPosEmb
        self.pos_emb = rel_pos_class(fmap_size, dim_head)

    def forward(self, fmap):
        heads, _, _, h, w = self.heads, *fmap.shape

        q, k, v = self.to_qkv(fmap).chunk(3, dim=1)
        q, k, v = map(lambda t: rearrange(t, 'b (h d) x y -> b h (x y) d', h=heads), (q, k, v))

        q = q * self.scale

        sim = einsum('b h i d, b h j d -> b h i j', q, k)
        sim += self.pos_emb(q)

        attn = sim.softmax(dim=-1)

        out = einsum('b h i j, b h j d -> b h i d', attn, v)
        out = rearrange(out, 'b h (x y) d -> b (h d) x y', x=h, y=w)
        return out


class BottleBlock(nn.Module):
    def __init__(
        self,
        *,
        dim,
        fmap_size,
        dim_out,
        proj_factor,
        downsample,
        heads=4,
        dim_head=128,
        rel_pos_emb=False,
        activation=nn.ReLU()
    ):
        super().__init__()

        # shortcut

        if dim != dim_out or downsample:
            kernel_size, stride, padding = (3, 2, 1) if downsample else (1, 1, 0)

            self.shortcut = nn.Sequential(
                nn.Conv2d(dim, dim_out, kernel_size, stride=stride, padding=padding, bias=False),
                nn.BatchNorm2d(dim_out),
                activation
            )
        else:
            self.shortcut = nn.Identity()

        # contraction and expansion

        attn_dim_in = dim_out // proj_factor
        attn_dim_out = heads * dim_head

        self.net = nn.Sequential(
            nn.Conv2d(dim, attn_dim_in, 1, bias=False),
            nn.BatchNorm2d(attn_dim_in),
            activation,
            Attention(
                dim=attn_dim_in,
                fmap_size=fmap_size,
                heads=heads,
                dim_head=dim_head,
                rel_pos_emb=rel_pos_emb
            ),
            nn.AvgPool2d((2, 2)) if downsample else nn.Identity(),
            nn.BatchNorm2d(attn_dim_out),
            activation,
            nn.Conv2d(attn_dim_out, dim_out, 1, bias=False),
            nn.BatchNorm2d(dim_out)
        )

        # init last batch norm gamma to zero

        nn.init.zeros_(self.net[-1].weight)

        # final activation

        self.activation = activation

    def forward(self, x):
        shortcut = self.shortcut(x)
        x = self.net(x)
        x += shortcut
        return self.activation(x)

# main bottle stack


class BottleStack(nn.Module):
    def __init__(
        self,
        *,
        dim,
        fmap_size,
        dim_out=2048,
        proj_factor=4,
        num_layers=3,
        heads=4,
        dim_head=128,
        downsample=True,
        rel_pos_emb=False,
        activation=nn.ReLU()
    ):
        super().__init__()
        fmap_size = pair(fmap_size)

        self.dim = dim
        self.fmap_size = fmap_size

        layers = []

        for i in range(num_layers):
            # pylint: disable=cell-var-from-loop
            is_first = i == 0
            dim = (dim if is_first else dim_out)
            layer_downsample = is_first and downsample

            layer_fmap_size = tuple(map(lambda t: t // (2 if downsample and not is_first else 1),
                                        fmap_size))

            layers.append(BottleBlock(
                dim=dim,
                fmap_size=layer_fmap_size,
                dim_out=dim_out,
                proj_factor=proj_factor,
                heads=heads,
                dim_head=dim_head,
                downsample=layer_downsample,
                rel_pos_emb=rel_pos_emb,
                activation=activation
            ))

        self.net = nn.Sequential(*layers)

    def forward(self, x):
        _, c, h, w = x.shape
        assert c == self.dim, \
            f'channels of feature map {c} must match channels given at init {self.dim}'
        assert h == self.fmap_size[0] and w == self.fmap_size[1], \
            f'height and width ({h} {w}) of feature map must match' \
            ' the fmap_size given at init {self.fmap_size}'
        return self.net(x)



"""Transforms for image pairs."""

from .blank_past import BlankPast, PreviousPast, RandomizeOneFrame
from .camera_shift import CameraShift
from .crop import Crop
from .encoders import Encoders
from .image_to_tracking import ImageToTracking
from .pad import Pad
from .sample_pairing import SamplePairing
from .single_image import SingleImage, Ungroup



"""
Interface for custom data.

This module handles datasets and is the class that you need to inherit from for your custom dataset.
This class gives you all the handles so that you can train with a new –dataset=mydataset.
The particular configuration of keypoints and skeleton is specified in the headmeta instances
"""


import argparse
import torch
import numpy as np
try:
    from pycocotools.coco import COCO
except ImportError:
    COCO = None

from openpifpaf.datasets import DataModule
from openpifpaf import encoder, headmeta, metric, transforms
from openpifpaf.datasets import collate_images_anns_meta, collate_images_targets_meta
from openpifpaf.plugins.coco import CocoDataset as CocoLoader

from .constants import get_constants, training_weights_local_centrality
from .metrics import MeanPixelError


class ApolloKp(DataModule):
    """
    DataModule for the Apollocar3d Dataset.
    """

    train_annotations = 'data-apollocar3d/annotations/apollo_keypoints_66_train.json'
    val_annotations = 'data-apollocar3d/annotations/apollo_keypoints_66_val.json'
    eval_annotations = val_annotations
    train_image_dir = 'data-apollocar3d/images/train/'
    val_image_dir = 'data-apollocar3d/images/val/'
    eval_image_dir = val_image_dir

    n_images = None
    square_edge = 513
    extended_scale = False
    orientation_invariant = 0.0
    blur = 0.0
    augmentation = True
    rescale_images = 1.0
    upsample_stride = 1
    min_kp_anns = 1
    b_min = 1  # 1 pixel

    eval_annotation_filter = True
    eval_long_edge = 0  # set to zero to deactivate rescaling
    eval_orientation_invariant = 0.0
    eval_extended_scale = False

    def __init__(self):
        super().__init__()
        if self.weights is not None:
            caf_weights = []
            for bone in self.CAR_SKELETON:
                caf_weights.append(max(self.weights[bone[0] - 1],
                                       self.weights[bone[1] - 1]))
            w_np = np.array(caf_weights)
            caf_weights = list(w_np / np.sum(w_np) * len(caf_weights))
        else:
            caf_weights = None
        cif = headmeta.Cif('cif', 'apollo',
                           keypoints=self.CAR_KEYPOINTS,
                           sigmas=self.CAR_SIGMAS,
                           pose=self.CAR_POSE,
                           draw_skeleton=self.CAR_SKELETON,
                           score_weights=self.CAR_SCORE_WEIGHTS,
                           training_weights=self.weights)
        caf = headmeta.Caf('caf', 'apollo',
                           keypoints=self.CAR_KEYPOINTS,
                           sigmas=self.CAR_SIGMAS,
                           pose=self.CAR_POSE,
                           skeleton=self.CAR_SKELETON,
                           training_weights=caf_weights)

        cif.upsample_stride = self.upsample_stride
        caf.upsample_stride = self.upsample_stride
        self.head_metas = [cif, caf]

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('data module Apollo')

        group.add_argument('--apollo-train-annotations',
                           default=cls.train_annotations)
        group.add_argument('--apollo-val-annotations',
                           default=cls.val_annotations)
        group.add_argument('--apollo-train-image-dir',
                           default=cls.train_image_dir)
        group.add_argument('--apollo-val-image-dir',
                           default=cls.val_image_dir)

        group.add_argument('--apollo-square-edge',
                           default=cls.square_edge, type=int,
                           help='square edge of input images')
        assert not cls.extended_scale
        group.add_argument('--apollo-extended-scale',
                           default=False, action='store_true',
                           help='augment with an extended scale range')
        group.add_argument('--apollo-orientation-invariant',
                           default=cls.orientation_invariant, type=float,
                           help='augment with random orientations')
        group.add_argument('--apollo-blur',
                           default=cls.blur, type=float,
                           help='augment with blur')
        assert cls.augmentation
        group.add_argument('--apollo-no-augmentation',
                           dest='apollo_augmentation',
                           default=True, action='store_false',
                           help='do not apply data augmentation')
        group.add_argument('--apollo-rescale-images',
                           default=cls.rescale_images, type=float,
                           help='overall rescale factor for images')
        group.add_argument('--apollo-upsample',
                           default=cls.upsample_stride, type=int,
                           help='head upsample stride')
        group.add_argument('--apollo-min-kp-anns',
                           default=cls.min_kp_anns, type=int,
                           help='filter images with fewer keypoint annotations')
        group.add_argument('--apollo-bmin',
                           default=cls.b_min, type=int,
                           help='b minimum in pixels')
        group.add_argument('--apollo-apply-local-centrality-weights',
                           dest='apollo_apply_local_centrality',
                           default=False, action='store_true',
                           help='Weigh the CIF and CAF head during training.')

        # evaluation
        assert cls.eval_annotation_filter
        group.add_argument('--apollo-no-eval-annotation-filter',
                           dest='apollo_eval_annotation_filter',
                           default=True, action='store_false')
        group.add_argument('--apollo-eval-long-edge', default=cls.eval_long_edge, type=int,
                           help='set to zero to deactivate rescaling')
        assert not cls.eval_extended_scale
        group.add_argument('--apollo-eval-extended-scale', default=False, action='store_true')
        group.add_argument('--apollo-eval-orientation-invariant',
                           default=cls.eval_orientation_invariant, type=float)
        group.add_argument('--apollo-use-24-kps', default=False, action='store_true',
                           help=('The ApolloCar3D dataset can '
                                 'be trained with 24 or 66 kps. If you want to train a model '
                                 'with 24 kps activate this flag. Change the annotations '
                                 'path to the json files with 24 kps.'))

    @classmethod
    def configure(cls, args: argparse.Namespace):
        # extract global information
        cls.debug = args.debug
        cls.pin_memory = args.pin_memory

        # Apollo specific
        cls.train_annotations = args.apollo_train_annotations
        cls.val_annotations = args.apollo_val_annotations
        cls.eval_annotations = cls.val_annotations
        cls.train_image_dir = args.apollo_train_image_dir
        cls.val_image_dir = args.apollo_val_image_dir
        cls.eval_image_dir = cls.val_image_dir

        cls.square_edge = args.apollo_square_edge
        cls.extended_scale = args.apollo_extended_scale
        cls.orientation_invariant = args.apollo_orientation_invariant
        cls.blur = args.apollo_blur
        cls.augmentation = args.apollo_augmentation  # loaded by the dest name
        cls.rescale_images = args.apollo_rescale_images
        cls.upsample_stride = args.apollo_upsample
        cls.min_kp_anns = args.apollo_min_kp_anns
        cls.b_min = args.apollo_bmin
        if args.apollo_use_24_kps:
            (cls.CAR_KEYPOINTS, cls.CAR_SKELETON, cls.HFLIP, cls.CAR_SIGMAS, cls.CAR_POSE,
             cls.CAR_CATEGORIES, cls.CAR_SCORE_WEIGHTS) = get_constants(24)
        else:
            (cls.CAR_KEYPOINTS, cls.CAR_SKELETON, cls.HFLIP, cls.CAR_SIGMAS, cls.CAR_POSE,
             cls.CAR_CATEGORIES, cls.CAR_SCORE_WEIGHTS) = get_constants(66)
        # evaluation
        cls.eval_annotation_filter = args.apollo_eval_annotation_filter
        cls.eval_long_edge = args.apollo_eval_long_edge
        cls.eval_orientation_invariant = args.apollo_eval_orientation_invariant
        cls.eval_extended_scale = args.apollo_eval_extended_scale
        if args.apollo_apply_local_centrality:
            if args.apollo_use_24_kps:
                raise Exception("Applying local centrality weights only works with 66 kps.")
            cls.weights = training_weights_local_centrality
        else:
            cls.weights = None

    def _preprocess(self):
        encoders = (encoder.Cif(self.head_metas[0], bmin=self.b_min),
                    encoder.Caf(self.head_metas[1], bmin=self.b_min))

        if not self.augmentation:
            return transforms.Compose([
                transforms.NormalizeAnnotations(),
                transforms.RescaleAbsolute(self.square_edge),
                transforms.CenterPad(self.square_edge),
                transforms.EVAL_TRANSFORM,
                transforms.Encoders(encoders),
            ])

        if self.extended_scale:
            rescale_t = transforms.RescaleRelative(
                scale_range=(0.2 * self.rescale_images,
                             2.0 * self.rescale_images),
                power_law=True, stretch_range=(0.75, 1.33))
        else:
            rescale_t = transforms.RescaleRelative(
                scale_range=(0.33 * self.rescale_images,
                             1.33 * self.rescale_images),
                power_law=True, stretch_range=(0.75, 1.33))

        return transforms.Compose([
            transforms.NormalizeAnnotations(),
            # transforms.AnnotationJitter(),
            transforms.RandomApply(transforms.HFlip(self.CAR_KEYPOINTS, self.HFLIP), 0.5),
            rescale_t,
            transforms.RandomApply(transforms.Blur(), self.blur),
            transforms.RandomChoice(
                [transforms.RotateBy90(),
                 transforms.RotateUniform(30.0)],
                [self.orientation_invariant, 0.2],
            ),
            transforms.Crop(self.square_edge, use_area_of_interest=True),
            transforms.CenterPad(self.square_edge),
            transforms.MinSize(min_side=32.0),
            transforms.TRAIN_TRANSFORM,
            transforms.Encoders(encoders),
        ])

    def train_loader(self):
        train_data = CocoLoader(
            image_dir=self.train_image_dir,
            ann_file=self.train_annotations,
            preprocess=self._preprocess(),
            annotation_filter=True,
            min_kp_anns=self.min_kp_anns,
            category_ids=[1],
        )
        return torch.utils.data.DataLoader(
            train_data, batch_size=self.batch_size, shuffle=not self.debug,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=collate_images_targets_meta)

    def val_loader(self):
        val_data = CocoLoader(
            image_dir=self.val_image_dir,
            ann_file=self.val_annotations,
            preprocess=self._preprocess(),
            annotation_filter=True,
            min_kp_anns=self.min_kp_anns,
            category_ids=[1],
        )
        return torch.utils.data.DataLoader(
            val_data, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=collate_images_targets_meta)

    @classmethod
    def common_eval_preprocess(cls):
        rescale_t = None
        if cls.eval_extended_scale:
            assert cls.eval_long_edge
            rescale_t = [
                transforms.DeterministicEqualChoice([
                    transforms.RescaleAbsolute(cls.eval_long_edge),
                    transforms.RescaleAbsolute((cls.eval_long_edge - 1) // 2 + 1),
                ], salt=1)
            ]
        elif cls.eval_long_edge:
            rescale_t = transforms.RescaleAbsolute(cls.eval_long_edge)

        if cls.batch_size == 1:
            padding_t = transforms.CenterPadTight(16)
        else:
            assert cls.eval_long_edge
            padding_t = transforms.CenterPad(cls.eval_long_edge)

        orientation_t = None
        if cls.eval_orientation_invariant:
            orientation_t = transforms.DeterministicEqualChoice([
                None,
                transforms.RotateBy90(fixed_angle=90),
                transforms.RotateBy90(fixed_angle=180),
                transforms.RotateBy90(fixed_angle=270),
            ], salt=3)

        return [
            transforms.NormalizeAnnotations(),
            rescale_t,
            padding_t,
            orientation_t,
        ]

    def _eval_preprocess(self):
        return transforms.Compose([
            *self.common_eval_preprocess(),
            transforms.ToAnnotations([
                transforms.ToKpAnnotations(
                    self.CAR_CATEGORIES,
                    keypoints_by_category={1: self.head_metas[0].keypoints},
                    skeleton_by_category={1: self.head_metas[1].skeleton},
                ),
                transforms.ToCrowdAnnotations(self.CAR_CATEGORIES),
            ]),
            transforms.EVAL_TRANSFORM,
        ])

    def eval_loader(self):
        eval_data = CocoLoader(
            image_dir=self.eval_image_dir,
            ann_file=self.eval_annotations,
            preprocess=self._eval_preprocess(),
            annotation_filter=self.eval_annotation_filter,
            min_kp_anns=self.min_kp_anns if self.eval_annotation_filter else 0,
            category_ids=[1] if self.eval_annotation_filter else [],
        )
        return torch.utils.data.DataLoader(
            eval_data, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=False,
            collate_fn=collate_images_anns_meta)

# TODO: make sure that 24kp flag is activated when evaluating a 24kp model
    def metrics(self):
        return [metric.Coco(
            COCO(self.eval_annotations),
            max_per_image=20,
            category_ids=[1],
            iou_type='keypoints',
            keypoint_oks_sigmas=self.CAR_SIGMAS
        ), MeanPixelError()]



import argparse
import numpy as np
import torch
import torchvision

import openpifpaf


class Cifar10(openpifpaf.datasets.DataModule):
    root_dir = 'data-cifar10/'
    download = False

    debug = False
    pin_memory = False

    def __init__(self):
        super().__init__()

        categories = ('plane', 'car', 'bird', 'cat', 'deer', 'dog', 'frog',
                      'horse', 'ship', 'truck')
        self.head_metas = [openpifpaf.headmeta.CifDet('cifdet', 'cifar10', categories)]

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('data module Cifar10')

        group.add_argument('--cifar10-root-dir',
                           default=cls.root_dir)
        assert not cls.download
        group.add_argument('--cifar10-download',
                           default=False, action='store_true')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        # extract global information
        cls.debug = args.debug
        cls.pin_memory = args.pin_memory

        # cocokp specific
        cls.root_dir = args.cifar10_root_dir
        cls.download = args.cifar10_download

    @staticmethod
    def _convert_data(parent_data, meta):
        image, category_id = parent_data

        anns = [{
            'bbox': np.asarray([5, 5, 21, 21], dtype=np.float32),
            'category_id': category_id + 1,
        }]

        return image, anns, meta

    def _preprocess(self):
        enc = openpifpaf.encoder.CifDet(self.head_metas[0])
        return openpifpaf.transforms.Compose([
            self._convert_data,
            openpifpaf.transforms.NormalizeAnnotations(),
            openpifpaf.transforms.CenterPadTight(16),
            openpifpaf.transforms.EVAL_TRANSFORM,
            openpifpaf.transforms.Encoders([enc]),
        ])

    def download_data(self):
        torchvision.datasets.CIFAR10(self.root_dir, download=True)

    def train_loader(self):
        train_data = openpifpaf.datasets.TorchDataset(
            torchvision.datasets.CIFAR10(self.root_dir, train=True, download=self.download),
            preprocess=self._preprocess(),
        )
        return torch.utils.data.DataLoader(
            train_data, batch_size=self.batch_size, shuffle=not self.debug,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=openpifpaf.datasets.collate_images_targets_meta)

    def val_loader(self):
        val_data = openpifpaf.datasets.TorchDataset(
            torchvision.datasets.CIFAR10(self.root_dir, train=False, download=self.download),
            preprocess=self._preprocess(),
        )
        return torch.utils.data.DataLoader(
            val_data, batch_size=self.batch_size, shuffle=not self.debug,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=openpifpaf.datasets.collate_images_targets_meta)

    def eval_loader(self):
        val_data = openpifpaf.datasets.TorchDataset(
            torchvision.datasets.CIFAR10(self.root_dir, train=False, download=self.download),
            preprocess=openpifpaf.transforms.Compose([
                self._convert_data,
                openpifpaf.transforms.NormalizeAnnotations(),
                openpifpaf.transforms.CenterPadTight(16),
                openpifpaf.transforms.EVAL_TRANSFORM,
                openpifpaf.transforms.ToAnnotations([
                    openpifpaf.transforms.ToDetAnnotations(
                        self.head_metas[0].categories),
                ]),
            ]),
        )
        return torch.utils.data.DataLoader(
            val_data, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=openpifpaf.datasets.collate_images_anns_meta)

    def metrics(self):
        return [openpifpaf.metric.Classification(
            categories=self.head_metas[0].categories,
        )]



import copy
import math
import numpy as np

from . import headmeta, utils


class Base:
    def inverse_transform(self, meta):
        raise NotImplementedError

    def json_data(self, coordinate_digits=2):
        raise NotImplementedError


class Annotation(Base):
    def __init__(self, keypoints, skeleton, sigmas=None, *,
                 categories=None, score_weights=None, suppress_score_index=None):
        self.keypoints = keypoints
        self.skeleton = skeleton
        self.sigmas = sigmas
        self.categories = categories
        self.score_weights = score_weights
        self.suppress_score_index = suppress_score_index

        self.category_id = 1
        self.data = np.zeros((len(keypoints), 3), dtype=np.float32)
        self.joint_scales = np.zeros((len(keypoints),), dtype=np.float32)
        self.fixed_score = None
        self.fixed_bbox = None
        self.decoding_order = []
        self.frontier_order = []

        self.skeleton_m1 = (np.asarray(skeleton) - 1).tolist()
        if score_weights is None:
            self.score_weights = np.ones((len(keypoints),))
        else:
            assert len(self.score_weights) == len(keypoints), "wrong number of scores"
            self.score_weights = np.asarray(self.score_weights)
        if self.suppress_score_index:
            self.score_weights[-len(self.suppress_score_index):] = 0.0
        self.score_weights /= np.sum(self.score_weights)

    @classmethod
    def from_cif_meta(cls, cif_meta: headmeta.Cif):
        scale = np.sqrt(
            (np.max(cif_meta.pose[:, 0]) - np.min(cif_meta.pose[:, 0]))
            * (np.max(cif_meta.pose[:, 1]) - np.min(cif_meta.pose[:, 1]))
        )
        ann = cls(keypoints=cif_meta.keypoints,
                  skeleton=cif_meta.draw_skeleton,
                  score_weights=cif_meta.score_weights)
        ann.set(cif_meta.pose, np.array(cif_meta.sigmas) * scale, fixed_score='')
        return ann

    @property
    def category(self):
        return self.categories[self.category_id - 1]

    def add(self, joint_i, xyv):
        self.data[joint_i] = xyv
        return self

    def set(self, data, joint_scales=None, *, category_id=1, fixed_score=None, fixed_bbox=None):
        """Set the data (keypoint locations, category, ...) for this instance."""
        self.data = data
        if joint_scales is not None:
            self.joint_scales = joint_scales
        else:
            self.joint_scales[:] = 0.0
            if self.sigmas is not None and fixed_bbox is not None:
                area = fixed_bbox[2] * fixed_bbox[3]
                self.joint_scales = np.sqrt(area) * np.asarray(self.sigmas)
        self.category_id = category_id
        self.fixed_score = fixed_score
        self.fixed_bbox = fixed_bbox
        return self

    def rescale(self, scale_factor):
        if len(scale_factor) == 2:
            scale_x, scale_y = scale_factor
            scale_factor = 0.5 * (scale_x + scale_y)
        else:
            scale_x = scale_factor
            scale_y = scale_factor

        self.data[:, 0] *= scale_x
        self.data[:, 1] *= scale_y
        if self.joint_scales is not None:
            self.joint_scales *= scale_factor
        for _, __, c1, c2 in self.decoding_order:
            c1[0:1] *= scale_x
            c1[1:2] *= scale_y
            c2[0:1] *= scale_x
            c2[1:2] *= scale_y
        return self

    @property
    def score(self):
        if self.fixed_score is not None:
            return self.fixed_score

        v = self.data[:, 2]
        if self.suppress_score_index is not None:
            v = np.copy(v)
            v[self.suppress_score_index] = 0.0
        # return 0.1 * np.max(v) + 0.9 * np.mean(np.square(v))
        # return np.mean(np.square(v))
        # return np.sum(self.score_weights * np.sort(np.square(v))[::-1])
        return np.sum(self.score_weights * np.sort(v)[::-1])

    def scale(self, v_th=0.5):
        m = self.data[:, 2] > v_th
        if not np.any(m):
            return 0.0
        return max(
            np.max(self.data[m, 0]) - np.min(self.data[m, 0]),
            np.max(self.data[m, 1]) - np.min(self.data[m, 1]),
        )

    def json_data(self, coordinate_digits=2):
        """Data ready for json dump."""

        # avoid visible keypoints becoming invisible due to rounding
        v_mask = self.data[:, 2] > 0.0
        keypoints = np.copy(self.data)
        keypoints[v_mask, 2] = np.maximum(0.01, keypoints[v_mask, 2])
        keypoints = np.around(keypoints.astype(np.float64), coordinate_digits)

        # convert to float64 before rounding because otherwise extra digits
        # will be added when converting to Python type
        data = {
            'keypoints': keypoints.reshape(-1).tolist(),
            'bbox': [round(float(c), coordinate_digits) for c in self.bbox()],
            'score': max(0.001, round(self.score, 3)),
            'category_id': self.category_id,
        }

        id_ = getattr(self, 'id_', None)
        if id_:
            data['id_'] = id_

        return data

    def bbox(self):
        if self.fixed_bbox is not None:
            return self.fixed_bbox
        return self.bbox_from_keypoints(self.data, self.joint_scales)

    @staticmethod
    def bbox_from_keypoints(kps, joint_scales):
        m = kps[:, 2] > 0
        if not np.any(m):
            return [0, 0, 0, 0]

        x = np.min(kps[:, 0][m] - joint_scales[m])
        y = np.min(kps[:, 1][m] - joint_scales[m])
        w = np.max(kps[:, 0][m] + joint_scales[m]) - x
        h = np.max(kps[:, 1][m] + joint_scales[m]) - y
        return [x, y, w, h]

    def inverse_transform(self, meta):
        ann = copy.deepcopy(self)

        # determine rotation parameters
        angle = -meta['rotation']['angle']
        rw = meta['rotation']['width']
        rh = meta['rotation']['height']
        cangle = math.cos(angle / 180.0 * math.pi)
        sangle = math.sin(angle / 180.0 * math.pi)

        # rotation
        if angle != 0.0:
            xy = ann.data[:, :2]
            x_old = xy[:, 0].copy() - (rw - 1) / 2
            y_old = xy[:, 1].copy() - (rh - 1) / 2
            xy[:, 0] = (rw - 1) / 2 + cangle * x_old + sangle * y_old
            xy[:, 1] = (rh - 1) / 2 - sangle * x_old + cangle * y_old
            if ann.fixed_bbox is not None:
                ann.fixed_bbox = utils.rotate_box(ann.fixed_bbox, rw - 1, rh - 1, angle)

        # offset
        ann.data[:, 0] += meta['offset'][0]
        ann.data[:, 1] += meta['offset'][1]
        if ann.fixed_bbox is not None:
            ann.fixed_bbox[:2] += meta['offset']

        # scale
        ann.data[:, 0] = ann.data[:, 0] / meta['scale'][0]
        ann.data[:, 1] = ann.data[:, 1] / meta['scale'][1]
        ann.joint_scales /= meta['scale'][0]
        if ann.fixed_bbox is not None:
            ann.fixed_bbox[:2] /= meta['scale']
            ann.fixed_bbox[2:] /= meta['scale']

        assert not np.any(np.isnan(ann.data))

        if meta['hflip']:
            w = meta['width_height'][0]
            ann.data[:, 0] = -ann.data[:, 0] + (w - 1)
            if meta.get('horizontal_swap'):
                ann.data[:] = meta['horizontal_swap'](ann.data)
            if ann.fixed_bbox is not None:
                ann.fixed_bbox[0] = -(ann.fixed_bbox[0] + ann.fixed_bbox[2]) - 1.0 + w

        for _, __, c1, c2 in ann.decoding_order:
            c1[:2] += meta['offset']
            c2[:2] += meta['offset']

            c1[:2] /= meta['scale']
            c2[:2] /= meta['scale']

        return ann


class AnnotationDet(Base):
    def __init__(self, categories):
        self.categories = categories
        self.category_id = None
        self.score = None
        self.bbox = None

    def set(self, category_id, score, bbox):
        """Set score to None for a ground truth annotation."""
        self.category_id = category_id
        self.score = score
        self.bbox = np.asarray(bbox)
        return self

    @property
    def category(self):
        return self.categories[self.category_id - 1]

    def json_data(self, coordinate_digits=2):
        return {
            'category_id': self.category_id,
            'category': self.category,
            'score': max(0.001, round(float(self.score), 3)),
            'bbox': [round(float(c), coordinate_digits) for c in self.bbox],
        }

    def inverse_transform(self, meta):
        ann = copy.deepcopy(self)

        angle = -meta['rotation']['angle']
        if angle != 0.0:
            rw = meta['rotation']['width']
            rh = meta['rotation']['height']
            ann.bbox = utils.rotate_box(ann.bbox, rw - 1, rh - 1, angle)

        ann.bbox[:2] += meta['offset']
        ann.bbox[:2] /= meta['scale']
        ann.bbox[2:] /= meta['scale']

        if meta['hflip']:
            w = meta['width_height'][0]
            ann.bbox[0] = -(ann.bbox[0] + ann.bbox[2]) - 1.0 + w

        return ann


class AnnotationCrowd(Base):
    def __init__(self, categories):
        self.categories = categories
        self.category_id = None
        self.bbox = None

    def set(self, category_id, bbox):
        """Set score to None for a ground truth annotation."""
        self.category_id = category_id
        self.bbox = np.asarray(bbox)
        return self

    @property
    def category(self):
        return self.categories[self.category_id - 1]

    def json_data(self, coordinate_digits=2):
        return {
            'category_id': self.category_id,
            'category': self.category,
            'bbox': [round(float(c), coordinate_digits) for c in self.bbox],
        }

    def inverse_transform(self, meta):
        ann = copy.deepcopy(self)

        angle = -meta['rotation']['angle']
        if angle != 0.0:
            rw = meta['rotation']['width']
            rh = meta['rotation']['height']
            ann.bbox = utils.rotate_box(ann.bbox, rw - 1, rh - 1, angle)

        ann.bbox[:2] += meta['offset']
        ann.bbox[:2] /= meta['scale']
        ann.bbox[2:] /= meta['scale']

        if meta['hflip']:
            w = meta['width_height'][0]
            ann.bbox[0] = -(ann.bbox[0] + ann.bbox[2]) - 1.0 + w

        return ann



import torch


def collate_images_anns_meta(batch):
    anns = [b[-2] for b in batch]
    metas = [b[-1] for b in batch]

    if len(batch[0]) == 4:
        # raw images are also in this batch
        images = [b[0] for b in batch]
        processed_images = torch.utils.data.dataloader.default_collate([b[1] for b in batch])
        return images, processed_images, anns, metas

    processed_images = torch.utils.data.dataloader.default_collate([b[0] for b in batch])
    return processed_images, anns, metas


def collate_images_targets_meta(batch):
    images = torch.utils.data.dataloader.default_collate([b[0] for b in batch])
    targets = torch.utils.data.dataloader.default_collate([b[1] for b in batch])
    metas = [b[2] for b in batch]
    return images, targets, metas


def collate_tracking_images_targets_meta(batch):
    images = torch.utils.data.dataloader.default_collate([
        im for group in batch for im in group[0]])

    targets = torch.utils.data.dataloader.default_collate([b[1] for b in batch])
    metas = [b[2] for b in batch]

    return images, targets, metas



import argparse
import logging
import time
from typing import List

import numpy as np
import torch

from .decoder import Decoder
from ..annotation import Annotation
from . import utils
from .. import headmeta, visualizer

LOG = logging.getLogger(__name__)


class CifCafDense(Decoder):
    dense_coupling = 0.0

    def __init__(self,
                 cif_meta: headmeta.Cif,
                 caf_meta: headmeta.Caf,
                 dense_caf_meta: headmeta.Caf):
        super().__init__()

        self.cif_meta = cif_meta
        self.caf_meta = caf_meta
        self.dense_caf_meta = dense_caf_meta

        # prefer decoders with more keypoints and associations
        self.priority += cif_meta.n_fields / 1000.0
        self.priority += caf_meta.n_fields / 1000.0
        self.priority += dense_caf_meta.n_fields / 1000.0

        # overwrite confidence scale
        self.dense_caf_meta.confidence_scales = [
            self.dense_coupling for _ in self.dense_caf_meta.skeleton
        ]

        concatenated_caf_meta = headmeta.Caf.concatenate(
            [caf_meta, dense_caf_meta])
        self.cifcaf = CifCaf([cif_meta], [concatenated_caf_meta])

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        """Command line interface (CLI) to extend argument parser."""
        group = parser.add_argument_group('CifCafDense decoder')
        group.add_argument('--dense-connections', nargs='?', type=float,
                           default=0.0, const=1.0)

    @classmethod
    def configure(cls, args: argparse.Namespace):
        """Take the parsed argument parser output and configure class variables."""
        cls.dense_coupling = args.dense_connections

    @classmethod
    def factory(cls, head_metas):
        if len(head_metas) < 3:
            return []
        if not cls.dense_coupling:
            return []
        return [
            CifCafDense(cif_meta, caf_meta, dense_meta)
            for cif_meta, caf_meta, dense_meta in zip(head_metas, head_metas[1:], head_metas[2:])
            if (isinstance(cif_meta, headmeta.Cif)
                and isinstance(caf_meta, headmeta.Caf)
                and isinstance(dense_meta, headmeta.Caf))
        ]

    def __call__(self, fields, initial_annotations=None):
        cifcaf_fields = [
            fields[self.cif_meta.head_index],
            torch.cat([
                fields[self.caf_meta.head_index],
                fields[self.dense_caf_meta.head_index],
            ], axis=0)
        ]
        return self.cifcaf(cifcaf_fields)


class CifCaf(Decoder):
    """Generate CifCaf poses from fields.

    :param: nms: set to None to switch off non-maximum suppression.
    """
    connection_method = 'blend'
    occupancy_visualizer = visualizer.Occupancy()
    nms_before_force_complete = False

    reverse_match = True

    def __init__(self,
                 cif_metas: List[headmeta.Cif],
                 caf_metas: List[headmeta.Caf],
                 *,
                 cif_visualizers=None,
                 cifhr_visualizers=None,
                 caf_visualizers=None):
        super().__init__()

        self.cif_metas = cif_metas
        self.caf_metas = caf_metas
        self.score_weights = cif_metas[0].score_weights
        self.confidence_scales = caf_metas[0].decoder_confidence_scales

        self.cif_visualizers = cif_visualizers
        if self.cif_visualizers is None:
            self.cif_visualizers = [visualizer.Cif(meta) for meta in cif_metas]
        self.cifhr_visualizers = cifhr_visualizers
        if self.cifhr_visualizers is None:
            self.cifhr_visualizers = [
                visualizer.CifHr(stride=meta.stride, field_names=meta.keypoints)
                for meta in cif_metas
            ]
        self.caf_visualizers = caf_visualizers
        if self.caf_visualizers is None:
            self.caf_visualizers = [visualizer.Caf(meta) for meta in caf_metas]

        self.cpp_decoder = torch.classes.openpifpaf_decoder.CifCaf(
            len(cif_metas[0].keypoints),
            torch.LongTensor(caf_metas[0].skeleton) - 1,
        )
        # prefer decoders with more keypoints and associations
        self.priority += sum(m.n_fields for m in cif_metas) / 1000.0
        self.priority += sum(m.n_fields for m in caf_metas) / 1000.0

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        """Command line interface (CLI) to extend argument parser."""
        CppCifCaf = torch.classes.openpifpaf_decoder.CifCaf  # pylint: disable=invalid-name

        group = parser.add_argument_group('CifCaf decoder')
        assert not CppCifCaf.get_force_complete()
        group.add_argument('--force-complete-pose',
                           default=False, action='store_true')
        group.add_argument('--force-complete-caf-th', type=float,
                           default=CppCifCaf.get_force_complete_caf_th(),
                           help='CAF threshold for force complete. Set to -1 to deactivate.')
        assert not cls.nms_before_force_complete
        group.add_argument('--nms-before-force-complete', default=False, action='store_true',
                           help='run an additional NMS before completing poses')

        assert utils.nms.Keypoints.get_keypoint_threshold() == CppCifCaf.get_keypoint_threshold()
        group.add_argument('--keypoint-threshold', type=float,
                           default=CppCifCaf.get_keypoint_threshold(),
                           help='filter keypoints by score')
        group.add_argument('--keypoint-threshold-rel', type=float,
                           default=CppCifCaf.get_keypoint_threshold_rel(),
                           help='filter keypoint connections by relative score')

        assert not CppCifCaf.get_greedy()
        group.add_argument('--greedy', default=False, action='store_true',
                           help='greedy decoding')
        group.add_argument('--connection-method',
                           default=cls.connection_method,
                           choices=('max', 'blend'),
                           help='connection method to use, max is faster')
        assert not CppCifCaf.get_block_joints()
        group.add_argument('--cifcaf-block-joints', default=False, action='store_true',
                           help='block joints')

        assert CppCifCaf.get_reverse_match()
        group.add_argument('--no-reverse-match',
                           default=True, dest='reverse_match', action='store_false')
        group.add_argument('--ablation-cifseeds-nms',
                           default=False, action='store_true')
        group.add_argument('--ablation-cifseeds-no-rescore',
                           default=False, action='store_true')
        group.add_argument('--ablation-caf-no-rescore',
                           default=False, action='store_true')
        group.add_argument('--ablation-independent-kp',
                           default=False, action='store_true')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        """Take the parsed argument parser output and configure class variables."""
        CppCifCaf = torch.classes.openpifpaf_decoder.CifCaf  # pylint: disable=invalid-name

        # force complete
        keypoint_threshold_nms = args.keypoint_threshold
        if args.force_complete_pose:
            if not args.ablation_independent_kp:
                args.keypoint_threshold = 0.0
            args.keypoint_threshold_rel = 0.0
            keypoint_threshold_nms = 0.0
        # check consistency
        if args.seed_threshold < args.keypoint_threshold:
            LOG.warning(
                'consistency: decreasing keypoint threshold to seed threshold of %f',
                args.seed_threshold,
            )
            args.keypoint_threshold = args.seed_threshold

        cls.nms_before_force_complete = args.nms_before_force_complete
        utils.nms.Keypoints.set_keypoint_threshold(keypoint_threshold_nms)

        CppCifCaf.set_force_complete(args.force_complete_pose)
        CppCifCaf.set_force_complete_caf_th(args.force_complete_caf_th)
        CppCifCaf.set_keypoint_threshold(args.keypoint_threshold)
        CppCifCaf.set_keypoint_threshold_rel(args.keypoint_threshold_rel)

        CppCifCaf.set_greedy(args.greedy)
        CppCifCaf.set_block_joints(args.cifcaf_block_joints)
        cls.connection_method = args.connection_method

        cls.reverse_match = args.reverse_match
        utils.CifSeeds.set_ablation_nms(args.ablation_cifseeds_nms)
        utils.CifSeeds.set_ablation_no_rescore(args.ablation_cifseeds_no_rescore)
        utils.CafScored.set_ablation_no_rescore(args.ablation_caf_no_rescore)
        if args.ablation_cifseeds_no_rescore and args.ablation_caf_no_rescore:
            utils.CifHr.set_ablation_skip(True)

    @classmethod
    def factory(cls, head_metas):
        if CifCafDense.dense_coupling:
            return []  # --dense-connections is requested, so use the other decoder
        return [
            CifCaf([meta], [meta_next])
            for meta, meta_next in zip(head_metas[:-1], head_metas[1:])
            if (isinstance(meta, headmeta.Cif)
                and isinstance(meta_next, headmeta.Caf))
        ]

    def __call__(self, fields, initial_annotations=None):
        if not initial_annotations:
            initial_annotations_t = None
            initial_ids_t = None
        else:
            initial_annotations_t = torch.empty(
                (len(initial_annotations), self.cif_metas[0].n_fields, 4))
            initial_ids_t = torch.empty((len(initial_annotations),), dtype=torch.int64)
            for i, (ann_py, ann_t) in enumerate(zip(initial_annotations, initial_annotations_t)):
                for f in range(len(ann_py.data)):
                    ann_t[f, 0] = float(ann_py.data[f, 2])
                    ann_t[f, 1] = float(ann_py.data[f, 0])
                    ann_t[f, 2] = float(ann_py.data[f, 1])
                    ann_t[f, 3] = float(ann_py.joint_scales[f])
                initial_ids_t[i] = getattr(ann_py, 'id_', -1)
            LOG.debug('initial annotations = %d', initial_annotations_t.size(0))

        for vis, meta in zip(self.cif_visualizers, self.cif_metas):
            vis.predicted(fields[meta.head_index])
        for vis, meta in zip(self.caf_visualizers, self.caf_metas):
            vis.predicted(fields[meta.head_index])

        start = time.perf_counter()
        annotations, annotation_ids = self.cpp_decoder.call_with_initial_annotations(
            fields[self.cif_metas[0].head_index],
            self.cif_metas[0].stride,
            fields[self.caf_metas[0].head_index],
            self.caf_metas[0].stride,
            initial_annotations_t,
            initial_ids_t,
        )
        LOG.debug('cpp annotations = %d (%.1fms)',
                  len(annotations),
                  (time.perf_counter() - start) * 1000.0)
        for vis in self.cifhr_visualizers:
            fields, low = self.cpp_decoder.get_cifhr()
            vis.predicted(fields, low)

        annotations_py = []
        for ann_data, ann_id in zip(annotations, annotation_ids):
            ann = Annotation(self.cif_metas[0].keypoints,
                             self.caf_metas[0].skeleton,
                             score_weights=self.score_weights)
            ann.data[:, :2] = ann_data[:, 1:3]
            ann.data[:, 2] = ann_data[:, 0]
            ann.joint_scales[:] = ann_data[:, 3]
            if ann_id != -1:
                ann.id_ = int(ann_id)
            annotations_py.append(ann)

        LOG.info('annotations %d: %s',
                 len(annotations_py),
                 [np.sum(ann.data[:, 2] > 0.1) for ann in annotations_py])
        return annotations_py



import argparse
import logging
import multiprocessing
import sys
import time
from typing import List

import torch

from .. import annotation, visualizer

LOG = logging.getLogger(__name__)


class DummyPool():
    @staticmethod
    def starmap(f, iterable):
        return [f(*i) for i in iterable]


class Decoder:
    """Generate predictions from image or field inputs.

    When creating a new generator, the main implementation goes into `__call__()`.
    """
    default_worker_pool = None
    torch_decoder = True

    def __init__(self):
        self.priority = 0.0  # reference priority for single image CifCaf
        self.worker_pool = self.default_worker_pool

        if self.worker_pool is None or self.worker_pool == 0:
            self.worker_pool = DummyPool()
        if isinstance(self.worker_pool, int):
            LOG.info('creating decoder worker pool with %d workers', self.worker_pool)
            assert not sys.platform.startswith('win'), (
                'not supported, use --decoder-workers=0 '
                'on windows'
            )

            # The new default for multiprocessing is 'spawn' for py38 on Mac.
            # This is not compatible with our configuration system.
            # For now, try to use 'fork'.
            # TODO: how to make configuration 'spawn' compatible
            multiprocessing_context = multiprocessing.get_context('fork')
            self.worker_pool = multiprocessing_context.Pool(self.worker_pool)

        self.last_decoder_time = 0.0
        self.last_nn_time = 0.0

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        """Command line interface (CLI) to extend argument parser."""

    @classmethod
    def configure(cls, args: argparse.Namespace):
        """Take the parsed argument parser output and configure class variables."""

    @classmethod
    def factory(cls, head_metas) -> List['Decoder']:
        """Create instances of an implementation."""
        raise NotImplementedError

    def __call__(self, fields, *, initial_annotations=None) -> List[annotation.Base]:
        """For single image, from fields to annotations."""
        raise NotImplementedError

    def __getstate__(self):
        return {
            k: v for k, v in self.__dict__.items()
            if k not in ('worker_pool',)
        }

    @classmethod
    def fields_batch(cls, model, image_batch, *, device=None):
        """From image batch to field batch."""
        start = time.time()

        def apply(f, items):
            """Apply f in a nested fashion to all items that are not list or tuple."""
            if items is None:
                return None
            if isinstance(items, (list, tuple)):
                return [apply(f, i) for i in items]
            return f(items)

        with torch.no_grad():
            if device is not None:
                image_batch = image_batch.to(device, non_blocking=True)

            with torch.autograd.profiler.record_function('model'):
                heads = model(image_batch)

            # to numpy
            with torch.autograd.profiler.record_function('tonumpy'):
                if cls.torch_decoder:
                    heads = apply(lambda x: x.cpu(), heads)
                else:
                    heads = apply(lambda x: x.cpu().numpy(), heads)

        # index by frame (item in batch)
        head_iter = apply(iter, heads)
        heads = []
        while True:
            try:
                heads.append(apply(next, head_iter))
            except StopIteration:
                break

        LOG.debug('nn processing time: %.1fms', (time.time() - start) * 1000.0)
        return heads

    def batch(self, model, image_batch, *, device=None, gt_anns_batch=None):
        """From image batch straight to annotations batch."""
        start_nn = time.perf_counter()
        fields_batch = self.fields_batch(model, image_batch, device=device)
        self.last_nn_time = time.perf_counter() - start_nn

        if gt_anns_batch is None:
            gt_anns_batch = [None for _ in fields_batch]

        if not isinstance(self.worker_pool, DummyPool):
            # remove debug_images to save time during pickle
            image_batch = [None for _ in fields_batch]
            gt_anns_batch = [None for _ in fields_batch]

        LOG.debug('parallel execution with worker %s', self.worker_pool)
        start_decoder = time.perf_counter()
        result = self.worker_pool.starmap(
            self._mappable_annotations, zip(fields_batch, image_batch, gt_anns_batch))
        self.last_decoder_time = time.perf_counter() - start_decoder

        LOG.debug('time: nn = %.1fms, dec = %.1fms',
                  self.last_nn_time * 1000.0,
                  self.last_decoder_time * 1000.0)
        return result

    def _mappable_annotations(self, fields, debug_image, gt_anns):
        if debug_image is not None:
            visualizer.Base.processed_image(debug_image)
        if gt_anns is not None:
            visualizer.Base.ground_truth(gt_anns)

        return self(fields)
